"""Provides the a class to facilitte easier plotting of Stoner Data.

Classes:
    PlotMixin:
        A class that uses matplotlib to plot data
"""

# pylint: disable=C0413
from __future__ import division

__all__ = ["PlotMixin"]
import os
from collections.abc import Mapping
from functools import wraps
import copy
from inspect import getfullargspec

import numpy as np
from scipy.interpolate import griddata

from matplotlib import pyplot as plt
from matplotlib import figure as mplfig
from matplotlib import cm, colors, colormaps

from Stoner.compat import string_types, index_types, int_types
from Stoner.tools import AttributeStore, isnone, isanynone, all_type, isiterable, typedList, get_option, fix_signature
from .formats import DefaultPlotStyle
from .utils import errorfill
from .utils import hsl2rgb


try:  # Check we've got 3D plotting
    from mpl_toolkits.mplot3d import Axes3D  # NOQA pylint: disable=unused-import
    from mpl_toolkits.axes_grid1 import host_subplot, inset_locator  # NOQA pylint: disable=unused-import
    import mpl_toolkits.axisartist as AA  # NOQA pylint: disable=unused-import

    _3D = True
except ImportError:
    _3D = False


def _getargspec(*args, **kargs):
    """Get the function signature spec."""
    ret = getfullargspec(*args, **kargs)
    if ret.args and ret.args[0] == "self":  # remove self for bound methods
        del ret.args[0]
    deflen = len(ret.defaults) if ret.defaults else 0
    kwargs = ret.args[-deflen:]
    kwargs.extend(ret.kwonlyargs)
    args = ret.args[: len(ret.args) - deflen]
    defaults = list(ret.defaults) if ret.defaults else []
    if ret.kwonlydefaults:
        defaults.extend(list(ret.kwonlydefaults.values()))
    return args, kwargs, defaults, len(kwargs) - len(ret.kwonlyargs)


FIG_KARGS = _getargspec(plt.figure)[1] + ["ax"]


def __mpl3DQuiver(x_coord, y_coord, z_coord, u_comp, v_comp, w_comp, **kargs):
    """Plot vector fields using mpltoolkit.quiver.

    Args:
        x_coord_coord_coord (array):
            x data coordinates
        y_coord (array):
            y data coordinates
        z_coord (array):
            z data coordinates
        u_comp (array):
            u data vector field component
        v_comp (array):
            v data vector field component
        w_comp (array):
            w data vector field component

    Return:
        matpltolib.pyplot.figure with a quiver plot.
    """
    if not _3D:
        raise RuntimeError("3D plotting Not available. Install matplotlib toolkits")
    if "ax" not in kargs:
        ax = plt.axes(projection="3d")
    else:
        ax = kargs["ax"]
    C = kargs.pop("color", None)
    if isinstance(C, np.ndarray) and C.ndim == 1:  # replace colours with a colour mapped array
        cmap = kargs.get("cmap", cm.viridis)
        C = cmap(C)
    vector_field = ax.quiver(x_coord, y_coord, z_coord, u_comp, v_comp, w_comp, colors=C, **kargs)

    return vector_field


class PlotMixin:
    r"""A mixin class that works with :py:class:`Stoner.Core.DataFile` to add additional plotting functionality.

    Args:
        args(tuple):
            Arguments to pass to :py:meth:`Stoner.Core.DataFile.__init__`
        kargs (dict):
            keyword arguments to pass to \b DataFile.__init__

    Attributes:
        ax (matplotlib.Axes):
            The current axes on the current figure.
        axes (list of matplotlib.Axes):
            A list of all the axes on the current figure
        fig (matplotlib.figure):
            The current figure object being worked with
        fignum (int):
            The current figure's number.
        labels (list of string):
            List of axis labels as aternates to the column_headers
        showfig (bool or None):
            Controls whether plot functions return a copy of the figure (True), the DataFile (False) or
            Nothing (None)
        subplots (list of matplotlib.Axes):
            essentially the same as :py:attr:`PlotMixin.axes` but ensures that the list of subplots is
            synchronised to the number fo Axes.
        template (:py:class:`Sonter.plot.formats.DefaultPlotStyle` or instance):
            A plot style template subclass or object that determines the format and appearance of plots.
    """

    positional_fmt = [plt.plot, plt.semilogx, plt.semilogy, plt.loglog]
    no_fmt = [errorfill]
    _template = DefaultPlotStyle()
    _legend = True
    multiple = "common"

    def __init__(self, *args, **kargs):  # Do the import of plt here to speed module load
        """Import plt and then calls the parent constructor."""
        self.__figure = None
        self._showfig = kargs.pop("showfig", True)  # Retains previous behaviour
        self._subplots = []
        self._public_attrs = {
            "fig": (int, mplfig.Figure),
            "labels": list,
            "template": (DefaultPlotStyle, type(DefaultPlotStyle)),
            "xlim": tuple,
            "ylim": tuple,
            "title": string_types,
            "xlabel": string_types,
            "ylabel": string_types,
            "_showfig": bool,
        }
        super().__init__(*args, **kargs)
        self._labels = typedList(string_types, [])
        if self.debug:
            print("Done PlotMixin init")

    # ===========================================================================================================
    # Properties of PlotMixin
    # ===========================================================================================================

    @property
    def ax(self):
        """Return the current axis number."""
        return self.axes.index(self.fig.gca())

    @ax.setter
    def ax(self, value):
        """Change the current axis number."""
        if isinstance(value, int) and 0 <= value < len(self.axes):
            self.fig.sca(self.axes[value])
        else:
            raise IndexError("Figure doesn't have enough axes")

    @property
    def axes(self):
        """Return  the current axes object."""
        if isinstance(self.__figure, mplfig.Figure):
            ret = self.__figure.axes
        else:
            ret = None
        return ret

    @property
    def cmap(self):  # pylint: disable=r0201
        """Get the current cmap."""
        return plt.get_cmap()

    @cmap.setter
    def cmap(self, cmap):  # pylint: disable=r0201
        """Set the plot cmap."""
        return plt.set_cmap(cmap)

    @property
    def fig(self):
        """Get the current figure."""
        return self.__figure

    @fig.setter
    def fig(self, value):
        """Set the current figure."""
        if isinstance(value, plt.Figure):
            self.__figure = self.template.new_figure(value.number)[0]
        elif isinstance(value, int):
            value = plt.figure(value)
            self.fig = value
        elif value is None:
            self.__figure = None
        else:
            raise NotImplementedError("fig should be a number of matplotlib figure")

    @property
    def fignum(self):
        """Return the current figure number."""
        return self.__figure.number

    @property
    def labels(self):
        """Return the labels for the plot columns."""
        if len(self._labels) == 0:
            return self.column_headers
        if len(self._labels) < len(self.column_headers):
            self._labels.extend(copy.deepcopy(self.column_headers[len(self._labels) :]))
        return self._labels

    @labels.setter
    def labels(self, value):
        """Set the labels for the plot columns."""
        if value is None:
            self._labels = typedList(string_types, self.column_headers)
        elif isiterable(value) and all_type(value, string_types):
            self._labels = typedList(string_types, value)
        else:
            raise TypeError(f"labels should be iterable and all strings, or None, not {type(value)}")

    @property
    def showfig(self):
        """Return either the current figure or self or None.

        The return value depends on whether the attribute is True or False or None.
        """
        if self._showfig is None or get_option("no_figs"):
            return None
        if self._showfig:
            return self.__figure
        return self

    @showfig.setter
    def showfig(self, value):
        """Force a figure to be displayed."""
        if not (value is None or isinstance(value, bool)):
            raise AttributeError(f"showfig should be a boolean value not a {type(value)}")
        self._showfig = value

    @property
    def subplots(self):
        """Return the subplot instances."""
        if self.__figure is not None and len(self.__figure.axes) > len(self._subplots):
            self._subplots = self.__figure.axes
        return self._subplots

    @property
    def template(self):
        """Return the current plot template."""
        if not hasattr(self, "_template"):
            self.template = DefaultPlotStyle()
        return self._template

    @template.setter
    def template(self, value):
        """Set the current template."""
        if isinstance(value, type) and issubclass(value, DefaultPlotStyle):
            value = value()
        if isinstance(value, DefaultPlotStyle):
            self._template = value
        else:
            raise ValueError(f"Template is not of the right class:{type(value)}")
        self._template.apply()

    def _Plot(self, ix, iy, fmt, plotter, figure, **kwords):
        """Private method for plotting a single plot to a figure.

        Args:
            ix (int):
                COlumn index of x data
            iy (int):
                Column index of y data
            fmt (str):
                Format of this plot
            plotter (callable):
                Routine used to plot this data
            figure (matplotlib.figure):
                Figure to plot data
            **kwords (dict):
                Other keyword arguments to pass through

        """
        kwords = copy.copy(kwords)  # Make sure we don;t mutate kwords by accident
        if "label" not in kwords:
            kwords["label"] = self._col_label(iy)
        x = self.column(ix)
        y = self.column(iy)
        mask = x.mask | y.mask
        x = x[~mask]
        y = y[~mask]
        for err in ["xerr", "yerr"]:  # Check whether we need to shorten errors too
            if err in kwords and len(kwords[err]) == len(mask):
                kwords[err] = kwords[err][~mask]
        if plotter in self.positional_fmt:  # plots with positional fmt
            if fmt is None:
                plotter(x, y, figure=figure, **kwords)
            else:
                plotter(x, y, fmt, figure=figure, **kwords)
        elif plotter in self.no_fmt:
            plotter(x, y, figure=figure, **kwords)
        else:
            if fmt is not None:
                kwords["fmt"] = fmt
            plotter(x, y, figure=figure, **kwords)
        for ax in figure.axes:
            self.template.customise_axes(ax, self)

    def _surface_plotter(self, x_coord, y_coord, z_coord, **kargs):
        """Plot a 3D color mapped surface.

        Args:
            x_coord, y_coord, z_coord (array):
                Data point coordinates
            kargs (dict):
                Other keywords to pass through

        ReturnsL
            A matplotib Figure

        This function attempts to work the same as the 2D surface plotter pcolor, but draws a 3D axes set
        """
        if not _3D:
            raise RuntimeError("3D plotting Not available. Install matplotlib toolkits")
        if not isinstance(self.__figure.gca(), Axes3D):
            ax = plt.axes(projection="3d")
        else:
            ax = self.__figure.gca()
        z_coord = np.nan_to_num(z_coord)
        surf = ax.plot_surface(x_coord, y_coord, z_coord, **kargs)
        self.fig.colorbar(surf, shrink=0.5, aspect=5, extend="both")

        return surf

    def _vector_color(self, xcol=None, ycol=None, ucol=None, vcol=None, wcol=None, **kargs):
        """Map a vector direction in the data to a value for use with a colormnap."""
        c = self._fix_cols(xcol=xcol, ycol=ycol, ucol=ucol, vcol=vcol, wcol=wcol, **kargs)

        if isinstance(c.wcol, index_types):  # 3D vector field
            wdata = self.column(c.wcol)
            phidata = (wdata - np.min(wdata)) / (np.max(wdata) - np.min(wdata))
        else:  # 2D vector field
            phidata = np.ones(len(self)) * 0.5
            wdata = phidata - 0.5
        qdata = 0.5 + (np.arctan2(self.column(c.ucol), self.column(c.vcol)) / (2 * np.pi))
        rdata = np.sqrt(self.column(c.ucol) ** 2 + self.column(c.vcol) ** 2 + wdata**2)
        rdata = rdata / rdata.max()
        Z = hsl2rgb(qdata, rdata, phidata).astype("f") / 255.0001 + 1e-7
        return Z

    def _span_slice(self, col, num):
        """Create a slice that covers the range of a given column."""
        v1, v2 = self.span(col)
        v = np.linspace(v1, v2, num)
        delta = v[1] - v[0]
        if isinstance(delta, int_types):
            v2 = v2 + delta
        else:
            v2 = v2 + delta / 2
        return slice(v1, v2, delta)

    def _vector_field_plotter(self, x_coord, y_coord, z_coord, u_comp, v_comp, w_comp, **kargs):
        """Plot vector fields using mayavi.mlab.

        Args:
            x_coord, y_coord, z_coord (array):
                Data point coordinates
            u_comp, v_comp, w_comp (array):
                U,V,W vector field component

        Returns:
            An mlab figure reference.
        """
        try:
            from mayavi import mlab  # might not work !
        except ImportError:
            return None
        if "scalars" in kargs:
            col_mode = "color_by_scalar"
        else:
            col_mode = "color_by_vector"
        if "scalars" in kargs and isinstance(kargs["scalars"], bool) and kargs["scalars"]:  # fancy mode on
            kargs["scalars"] = np.arange(len(self))
            colors = self._vector_color() * 255
            colors = np.column_stack((colors, np.ones(len(self)) * 255))
            quiv = mlab.quiver3d(x_coord, y_coord, z_coord, u_comp, v_comp, w_comp, **kargs)
            quiv.glyph.color_mode = col_mode
            quiv.module_manager.scalar_lut_manager.lut.table = colors
        else:
            quiv = mlab.quiver3d(x_coord, y_coord, z_coord, u_comp, v_comp, w_comp, **kargs)
            quiv.glyph.color_mode = col_mode
        return quiv

    def _col_label(self, index, join=False):
        """Look up a column and see if it exists in self._lables, otherwise get from self.column_headers.

        Args:
            index (column index type):
                Column to return label for

        Returns:
            String type representing the column label.
        """
        ix = self.find_col(index)
        if isinstance(ix, list):
            if join:
                return ",".join([self._col_label(i) for i in ix])
            return [self._col_label(i) for i in ix]
        return self.labels[ix]

    def __dir__(self):
        """Handle the local attributes as well as the inherited ones."""
        attr = dir(type(self))
        attr.extend(super().__dir__())
        attr.extend(list(self.__dict__.keys()))
        attr.extend(["fig", "axes", "labels", "subplots", "template"])
        attr.extend(("xlabel", "ylabel", "title", "xlim", "ylim"))
        attr = list(set(attr))
        return sorted(attr)

    def _fix_cols(self, scalar=True, **kargs):
        """Sorts out axis specs, replacing with contents from setas as necessary."""
        startx = kargs.pop("startx", 0)

        c = self.setas._get_cols(startx=startx)
        for k in "xcol", "xerr":
            if k in kargs and kargs[k] is None:
                kargs[k] = c[k]
        for k in ["ycol", "zcol", "ucol", "vcol", "wcol", "yerr", "zerr"]:
            if k in kargs and kargs[k] is None and isiterable(c[k]) and len(c[k]) > 0:
                if kargs.get("multi_y", not scalar):
                    kargs[k] = c[k]
                else:
                    kargs[k] = c[k][0]
            elif k in c and k in kargs and kargs[k] is None:
                kargs[k] = c[k]
        for k in list(kargs.keys()):
            if k not in ["xcol", "ycol", "zcol", "ucol", "vcol", "wcol", "xerr", "yerr", "zerr"]:
                del kargs[k]
        return AttributeStore(kargs)

    def _fix_fig(self, figure, **kargs):
        """Sorts out the matplotlib figure handling."""
        if isinstance(figure, bool) and not figure:
            figure, ax = self.template.new_figure(None, **kargs)
        elif not isinstance(figure, bool) and isinstance(figure, int):
            figure, ax = self.template.new_figure(figure, **kargs)
        elif isinstance(figure, mplfig.Figure):
            figure, ax = self.template.new_figure(figure.number, **kargs)
        elif isinstance(self.__figure, mplfig.Figure):
            figure = self.__figure
            ax = self.__figure.gca(**kargs)
        else:
            figure, ax = self.template.new_figure(None, **kargs)
        self.__figure = figure
        figure.sca(ax)  # Esur4e we're set for plotting on the correct axes
        return figure, ax

    def _fix_kargs(self, function=None, defaults=None, otherkargs=None, **kargs):
        """Fix parameters to the plotting function to provide defaults and no extransous arguments.

        Returns:
            dictionary of correct arguments, dictionary of all arguments,dictionary of keyword arguments
        """
        if defaults is None:
            defaults = dict()
        defaults.update(kargs)

        pass_fig_kargs = {}
        for k in set(FIG_KARGS) & set(kargs.keys()):
            pass_fig_kargs[k] = kargs[k]
            if k not in otherkargs and k not in defaults:
                del kargs[k]

        # Defaults now a dictionary of default arguments overlaid with keyword argument values
        # Now inspect the plotting function to see what it takes.
        if function is None:
            function = defaults["plotter"]
        if isinstance(function, string_types):
            if "projection" in kargs:
                projection = kargs["projection"]
            else:
                projection = "rectilinear"
            function = plt.axes(projection=projection).__getattribute__(function)
            if self.__figure is not plt.gcf():
                plt.close(plt.gcf())

        (args, kwargs) = _getargspec(function)[:2]
        # Manually override the list of arguments that the plotting function takes if it takes keyword dictionary
        if isinstance(otherkargs, (list, tuple)):
            kwargs.extend(otherkargs)
        nonkargs = {}
        func_kwargs = {}
        for key, value in defaults.items():
            if key in kwargs:
                func_kwargs[key] = value
            else:
                nonkargs[key] = value
        return func_kwargs, nonkargs, pass_fig_kargs

    def _fix_titles(self, ix, multiple, **kargs):
        """Do the titling and labelling for a matplotlib plot."""
        self.template.annotate(ix, multiple, self, **kargs)
        if "show_plot" in kargs and kargs["show_plot"]:
            plt.ion()
            plt.draw()
            plt.show()
        if "save_filename" in kargs and kargs["save_filename"] is not None:
            plt.savefig(str(kargs["save_filename"]))

    def __getattr__(self, name):
        """Wrap attribute access with extra renaming logic.

        Args:
            name (string):
                Name of attribute the following attributes are supported:
                    -   fig - the current plt figure reference
                    -   axes - the plt axes object for the current plot
                    -   xlim - the X axis limits
                    -   ylim - the Y axis limits

                    All other attributes are passed over to the parent class
        """
        func = None
        o_name = name
        mapping = dict(
            [
                ("plt_", (plt, "pyplot")),  # Need to be explicit in 2.7!
                ("ax_", (plt.Axes, "axes")),
                ("fig_", (plt.Figure, "figure")),
            ]
        )

        try:
            return object.__getattribute__(self, o_name)
        except AttributeError:
            pass

        if plt.get_fignums():  # Store the current figure and axes
            tfig = plt.gcf()
            tax = tfig.gca()  # protect the current axes and figure
        else:
            tfig = None
            tax = None

        # First look for a function in the pyplot package
        for prefix, (obj, key) in mapping.items():
            name = o_name[len(prefix) :] if o_name.startswith(prefix) else o_name
            if name in dir(obj):
                try:
                    return self._pyplot_proxy(name, key)
                except AttributeError:
                    pass

        # Nowcheck for prefixed access on axes and figures with get_
        if name.startswith("ax_") and f"get_{name[3:]}" in dir(plt.Axes):
            name = name[3:]
        if f"get_{name}" in dir(plt.Axes) and self.__figure:
            ax = self.fig.gca()
            func = ax.__getattribute__(f"get_{name}")
        if name.startswith("fig_") and f"get_{name[4:]}" in dir(plt.Figure):
            name = name[4:]
        if f"get_{name}" in dir(plt.Figure) and self.__figure:
            fig = self.fig
            func = fig.__getattribute__(f"get_{name}")

        if func is None:  # Ok Fallback to lookinf again at parent class
            return super().__getattr__(o_name)

        # If we're still here then we're calling a proxy from that figure or axes
        ret = func()
        if tfig is not None:  # Restore the current figures
            plt.figure(tfig.number)
            plt.sca(tax)
        return ret

    def _pyplot_proxy(self, name, what):
        """Proxy for accessing :py:module:`matplotlib.pyplot` functions."""
        if what not in ["axes", "figure", "pyplot"]:
            raise SyntaxError("pyplot proxy can't figure out what to get proxy from.")
        if what == "pyplot":
            obj = plt
        elif what == "figure" and self.__figure:
            obj = self.__figure
        elif what == "axes" and self.__figure:
            obj = self.__figure.gca()
        else:
            raise AttributeError(
                "Attempting to manipulate the methods on a figure or axes before a"
                + " figure has been created for this Data."
            )
        func = getattr(obj, name)

        if not callable(func):  # Bug out if this isn't a callable proxy!
            return func

        @wraps(func)
        def _proxy(*args, **kargs):
            ret = func(*args, **kargs)
            return ret

        return fix_signature(_proxy, func)

    def __setattr__(self, name, value):
        """Set the specified attribute.

        Args:
            name (string):
                The name of the attribute to set. The cuirrent attributes are supported:
                    -   fig - set the plt figure instance to use
                    -   xlabel - set the X axis label text
                    -   ylabel - set the Y axis label text
                    -   title - set the plot title
                    -   subtitle - set the plot subtitle
                    -   xlim - set the x-axis limits
                    -   ylim - set the y-axis limits

                Only "fig" is supported in this class - everything else drops through to the parent class
                value (any): The value of the attribute to set.
        """
        if plt.get_fignums():
            tfig = plt.gcf()
            if len(tfig.axes):
                tax = tfig.gca()  # protect the current axes and figure
            else:
                tax = None
        else:
            tfig = None
            tax = None
        func = None
        o_name = name
        if name.startswith("ax_") and f"set_{name[3:]}" in dir(plt.Axes):
            name = name[3:]
        if f"set_{name}" in dir(plt.Axes) and self.fig:
            if self.fig is None:  # oops we need a figure first!
                self.figure()
            ax = self.fig.gca()
            func = ax.__getattribute__(f"set_{name}")
        elif name.startswith("fig_") and f"set_{name[4:]}" in dir(plt.Figure):
            name = f"set_{name[4:]}"
            func = getattr(self.fig, name)
        elif f"set_{name}" in dir(plt.Figure) and self.fig:
            if self.fig is None:  # oops we need a figure first!
                self.figure()
            fig = self.fig
            func = fig.__getattribute__(f"set_{name}")

        if o_name in dir(type(self)) or func is None:
            try:
                return super().__setattr__(o_name, value)
            except AttributeError:
                pass

        if func is None:
            raise AttributeError(f"Unable to set attribute {o_name}")

        if isinstance(value, string_types) and "$" not in value:
            value = value.format(**self)

        if not isiterable(value) or isinstance(value, string_types):
            value = (value,)
        if isinstance(value, Mapping):
            func(**value)
        else:
            func(*value)
        if tfig is not None:
            plt.figure(tfig.number)
            if tax is not None:
                plt.sca(tax)

    def add_column(self, column_data, header=None, index=None, **kargs):
        """Append a column of data or inserts a column to a datafile instance.

        Args:
            column_data (:py:class:`numpy.array` or list or callable):
                Data to append or insert or a callable function that will generate new data

        Keyword Arguments:
            column_header (string):
                The text to set the column header to, if not supplied then defaults to 'col#'
            index (int or string):
                The  index (numeric or string) to insert (or replace) the data
            func_args (dict):
                If column_data is a callable object, then this argument can be used to supply a dictionary of function
                arguments to the callable object.
            replace (bool):
                Replace the data or insert the data (default)
            setas (str):
                Set the type of column (x,y,z data etc - see :py:attr:`Stoner.Core.DataFile.setas`)

        Returns:
            A :py:class:`DataFile` instance with the additional column inserted.

        Note:
            Like most :py:class:`DataFile` methods, this method operates in-place in that it also modifies
            the original DataFile Instance as well as returning it.
        """
        # Call the parent method and then update this label
        super().add_column(column_data, header=header, index=index, **kargs)
        # Mostly this is duplicating the parent method
        if index is None:
            index = len(self.column_headers) - 1
        else:
            index = self.find_col(index)

        self.labels = (
            self.labels[:index]
            + self.column_headers[index : len(self.column_headers) - len(self.labels) + index]
            + self.labels[index:]
        )
        return self

    def colormap_xyz(self, xcol=None, ycol=None, zcol=None, **kargs):
        """Make a xyz plot that forces the use of plt.colormap.

        Args:
            xcol (index):
                Xcolumn index or label
            ycol (index):
                Y column index or label
            zcol (index):
                Z column index or label

        Keyword Arguments:
            shape (two-tuple):
                Number of points along x and y in the grid - defaults to a square of sidelength = square root of the
                length of the data.
            xlim (tuple):
                The xlimits, defaults to automatically determined from data
            ylim (tuple):
                The ylimits, defaults to automatically determined from data
            plotter (function):
                Function to use to plot data. Defaults to plt.contour
            colorbar (bool):
                Draw the z-scale color bar beside the plot (True by default)
            show_plot (bool):
                Turn on interfactive plotting and show plot when drawn
            save_filename (string or None):
                If set to a string, save the plot with this filename
            figure (integer or matplotlib.figure or boolean):
                Controls which figure is used for the plot, or if a new figure is opened.
            **kargs (dict):
                Other arguments are passed on to the plotter.

        Returns:
            A matplotlib figure
        """
        kargs["plotter"] = kargs.get("plotter", plt.pcolor)
        kargs["projection"] = kargs.get("projection", "rectilinear")
        xlim = kargs.pop("xlim", None)
        ylim = kargs.pop("ylim", None)
        shape = kargs.pop("shape", None)
        colorbar = kargs.pop("colorbar", True)
        ax = self.plot_xyz(xcol, ycol, zcol, shape, xlim, ylim, **kargs)
        if colorbar:
            plt.colorbar()
        return ax

    def contour_xyz(self, xcol=None, ycol=None, zcol=None, shape=None, xlim=None, ylim=None, plotter=None, **kargs):
        """Make a xyz plot that forces the use of plt.contour.

         Args:
            xcol (index):
                Xcolumn index or label
            ycol (index):
                Y column index or label
            zcol (index):
                Z column index or label

        Keyword Arguments:
            shape (two-tuple):
                Number of points along x and y in the grid - defaults to a square of sidelength = square root of the
                length of the data.
            xlim (tuple):
                The xlimits, defaults to automatically determined from data
            ylim (tuple):
                The ylimits, defaults to automatically determined from data
            plotter (function):
                Function to use to plot data. Defaults to plt.contour
            show_plot (bool):
                Turn on interfactive plotting and show plot when drawn
            save_filename (string or None):
                If set to a string, save the plot with this filename
            figure (integer or matplotlib.figure or boolean):
                Controls which figure is used for the plot, or if a new figure is opened.
            **kargs (dict):
                Other arguments are passed on to the plotter.

        Returns:
            A matplotlib figure
        """
        if plotter is None:
            plotter = plt.contour
        kargs["plotter"] = plotter
        return self.plot_xyz(xcol, ycol, zcol, shape, xlim, ylim, **kargs)

    def figure(self, figure=None, projection="rectilinear", **kargs):
        """Set the figure used by :py:class:`Stoner.plot.PlotMixin`.

        Args:
            figure (matplotlib.Figure or int):
                Figure to switch to

        Returns:
            The current Stoner.plot.PlotMixin instance
        """
        if figure is None:
            figure = self.template.new_figure(None, projection=projection, **kargs)[0]
        elif isinstance(figure, int):
            figure = self.template.new_figure(figure, projection=projection, **kargs)[0]
        elif isinstance(figure, mplfig.Figure):
            figure = self.template.new_figure(figure.number, projection=projection, **kargs)[0]
        self.__figure = figure
        return self

    def griddata(
        self,
        xcol=None,
        ycol=None,
        zcol=None,
        ucol=None,
        shape=None,
        xlim=None,
        ylim=None,
        zlim=None,
        method="linear",
        **kargs,
    ):
        """Convert xyz data onto a regular grid.

        Args:
            xcol (index):
                Xcolumn index or label
            ycol (index):
                Y column index or label
            zcol (index):
                Z column index or label

        Keyword Arguments:
            ucol (index):
                U (magnitude) column index or label
            shape (two-tuple, three-ruple):
                Number of points along x and y in the grid - defaults to a square of sidelength = square root of
                the length of the data.
            xlim (tuple):
                The xlimits
            ylim (tuple):
                The ylimits
            zlim (tuple):
                The ylimits
            method (string):
                Type of interpolation to use, default is linear

            ReturnsL
                (X,Y,Z) or (X,Y,Z,M):
                    three two dimensional arrays of the coordinates of the interpolated data or 4 three diemensional
                    arrays of the interpolated data

        Notes:
            Depending on whether 3 or 4 columns of data can be identified, this method will produce data for a
            :math:`Z(X,Y)` plot or a :math:`M(X,Y,Z)` volumetric plot.
        """
        startx = kargs.pop("startx", 0)
        cols = self.setas._get_cols(startx=startx)
        if isanynone(xcol, ycol, zcol):
            if xcol is None:
                xcol = cols["xcol"]
            if ycol is None:
                ycol = cols["ycol"][0]
            if zcol is None:
                if len(cols["zcol"]) > 0:
                    zcol = cols["zcol"][0]
        if ucol is None and len(cols["ucol"]) > 0:
            ucol = cols["ucol"][0]

        dims = 3 if cols["axes"] == 3 and cols["has_ucol"] else 2
        if getattr(zcol, "size", 0) != 0:
            shape = [np.unique(self.column(x)).size for x in [xcol, ycol]]
        else:
            shape = [np.unique(self.column(x)).size for x in [xcol, ycol, zcol]]

        lims = [xlim, ylim, zlim]
        extents = [xlim, ylim, zlim]
        for dim, (lim, col) in enumerate(zip(lims, [xcol, ycol, zcol])):
            if dim >= dims:  # Don;t bother analysing more dimensions than we have
                break
            if lim is None:
                lim = self._span_slice(col, shape[dim])
            elif isinstance(lim, tuple) and len(lim) > 1:
                lim = lim + ((lim[1] - lim[0]) / shape[dim],)
                lim = slice(lim[0], lim[1], lim[2])
            else:
                raise RuntimeError(f"{'xyz'[dim]} limit speciifcation not understood")
            lims[dim] = lim
            extents[dim] = slice(lim.start - lim.step / 2, lim.stop + lim.step / 2, lim.step)

        xlim = lims[0]
        ylim = lims[1]
        zlim = lims[2]

        if dims == 2:
            pts = np.mgrid[xlim, ylim].T
            points = np.array([self.column(xcol), self.column(ycol)]).T
            if zcol is None:
                zdata = np.zeros(len(self))
            elif isinstance(zcol, np.ndarray) and zcol.shape[0] == len(self):  # zcol is some data
                zdata = zcol
            else:
                zdata = self.column(zcol)
            if zdata.ndim == 1:
                Z = griddata(points, zdata, pts, method=method)
            elif zdata.ndim == 2:
                Z = np.zeros((pts.shape[0], pts.shape[1], zdata.shape[1]))
                for i in range(zdata.shape[1]):
                    Z[:, :, i] = griddata(points, zdata[:, i], pts, method=method)

            return pts[:, :, 0], pts[:, :, 1], Z
        elif dims == 3:
            pts = np.mgrid[xlim, ylim, zlim].T
            vpts = np.mgrid[extents[0], extents[1], extents[2]].T

            points = np.array([self.column(xcol), self.column(ycol), self.column(zcol)]).T
            if ucol is None:
                udata = np.zeros(len(self))
            elif isinstance(ucol, np.ndarray) and ucol.shape[0] == len(self):  # zcol is some data
                udata = ucol
            else:
                udata = self.column(ucol)
            if udata.ndim == 1:
                U = griddata(points, udata, pts, method=method)
            elif udata.ndim == 2:
                U = np.zeros((pts.shape[0], pts.shape[1], udata.shape[1]))
                for i in range(udata.shape[1]):
                    U[:, :, i] = griddata(points, udata[:, i], pts, method=method)

            return vpts[:, :, :, 0], vpts[:, :, :, 1], vpts[:, :, :, 2], U

    def image_plot(self, xcol=None, ycol=None, zcol=None, shape=None, xlim=None, ylim=None, **kargs):
        """Grid up the three columns of data and plot.

        Args:
            xcol (index):
                Xcolumn index or label
            ycol (index):
                Y column index or label
            zcol (index):
                Z column index or label

        Keyword Arguments:
            shape (two-tuple):
                Number of points along x and y in the grid - defaults to a square of sidelength = square root of the
                length of the data.
            xlim (tuple):
                The xlimits, defaults to automatically determined from data
            ylim (tuple):
                The ylimits, defaults to automatically determined from data
            xlabel (string):
                X axes label. Default is None - guess from xvals or metadata
            ylabel (string):
                Y axes label, Default is None - guess from metadata
            zlabel (string):
                Z axis label, Default is None - guess from metadata
            plotter (function):
                Function to use to plot data. Defaults to plt.contour
            show_plot (bool):
                Turn on interfactive plotting and show plot when drawn
            save_filename (string or None):
                If set to a string, save the plot with this filename
            figure (integer or matplotlib.figure or boolean):
                Controls which figure is used for the plot, or if a new figure is opened.
            **kargs (dict):
                Other arguments are passed on to the plotter.

        Returns:
            A matplotlib figure
        """
        locals().update(self._fix_cols(xcol=xcol, ycol=ycol, **kargs))

        X, Y, Z = self.griddata(xcol, ycol, zcol, shape, xlim, ylim)
        defaults = {
            "origin": "lower",
            "interpolation": "bilinear",
            "plotter": plt.imshow,
            "title": self.filename,
            "cmap": cm.jet,
            "figure": self.__figure,
            "xlabel": self._col_label(self.find_col(xcol)),
            "ylabel": self._col_label(self.find_col(ycol)),
            "extents": [self.x.min(), self.x.max(), self.y.min(), self.y.max()],
        }
        kargs, nonkargs, _ = self._fix_kargs(None, defaults, **kargs)
        plotter = nonkargs["plotter"]
        self.__figure = self._fix_fig(nonkargs["figure"])[0]
        if "cmap" in kargs and isinstance(kargs["cmap"], str):
            cmap = colormaps[kargs["cmap"]]
        elif "cmap" in nonkargs and isinstance(kargs["cmap"], str):
            cmap = colormaps(nonkargs["cmap"])
        if Z.ndim == 2:
            Z = cmap(Z)
        elif Z.ndim != 3:
            raise RuntimeError(f"Z Data has a bad shape: {Z.shape}")
        xmin = np.min(X.ravel())
        xmax = np.max(X.ravel())
        ymin = np.min(Y.ravel())
        ymax = np.max(Y.ravel())
        extent = [xmin, xmax, ymin, ymax]
        fig = plotter(Z, extent=extent, aspect="auto", **kargs)
        self._fix_titles(0, "none", **nonkargs)
        return fig

    def inset(self, parent=None, loc=None, width=0.35, height=0.30, **kargs):  # pylint: disable=r0201
        """Add a new set of axes as an inset to the current plot.

        Keyword Arguments:
            parent (matplotlib axes):
                Which set of axes to add inset to, defaults to the current set
            loc (int or string):
                Inset location - can be a string like *top right* or *upper right* or a number.
            width,height (int,float or string):
                the dimensions of the inset specified as a integer %, or floating point fraction of the parent axes,
                or as a string measurement.
            kargs (dictionary):
                all other keywords are passed through to inset_locator.inset_axes

        Returns:
            A new set of axes
        """
        locations = [
            "best",
            "upper right",
            "upper left",
            "lower left",
            "lower right",
            "right",
            "center left",
            "center right",
            "lower center",
            "upper center",
            "center",
        ]
        locations2 = [
            "best",
            "top right",
            "top left",
            "bottom left",
            "bottom right",
            "outside",
            "leftside",
            "rightside",
            "bottom",
            "top",
            "middle",
        ]
        if isinstance(loc, string_types):
            if loc in locations:
                loc = locations.index(loc)
            elif loc in locations2:
                loc = locations2.index(loc)
            else:
                raise RuntimeError(f"Couldn't work out where {loc} was supposed to be")
        if isinstance(width, int):
            width = f"{width}%"
        elif isinstance(width, float) and 0 < width <= 1:
            width = f"{width*100}%"
        elif not isinstance(width, string_types):
            raise RuntimeError(f"didn't Recognize width specification {width}")
        if isinstance(height, int):
            height = f"{height}%"
        elif isinstance(height, float) and 0 < height <= 1:
            height = "{height * 100}%"
        elif not isinstance(height, string_types):
            raise RuntimeError("didn't Recognize height specification {height}")
        if parent is None:
            parent = plt.gca()
        return inset_locator.inset_axes(parent, width, height, loc, **kargs)

    def legend(self, *args, **kargs):
        """Pass Through to stop attribute access over-riding a handy method."""
        self.gca().legend(*args, **kargs)
        return self

    def plot(self, *args, **kargs):
        """Try to make an appropriate plot based on the defined column assignments.

        The column assignments are examined to determine whether to plot and x,y plot or an x,y,z plot
        and whether to plot error bars (for an x,y plot). All keyword argume nts are passed through to
        the selected plotting routine.
        """
        if len(args) != 0:
            axes = len(args)
        else:
            _ = self._col_args(**kargs)
            axes = _.axes
        if "template" in kargs:
            self.template = kargs.pop("template")

        if axes == 3 and ("ucol" in args or _.has_ucol):
            axes = 7  # trick to allow voxel plot for xyzu

        plotters = {
            2: self.plot_xy,
            3: self.plot_xyz,
            4: self.plot_xyuv,
            5: self.plot_xyuv,
            6: self.plot_xyzuvw,
            7: self.plot_voxels,
        }
        try:
            plotter = plotters.get(axes, None)
            ret = plotter(*args, **kargs)
            plt.show()
        except KeyError as err:
            raise RuntimeError("Unable to work out plot type !") from err
        return ret

    def plot_matrix(
        self,
        xvals=None,
        yvals=None,
        rectang=None,
        cmap=plt.cm.plasma,
        show_plot=True,
        title="",
        xlabel=None,
        ylabel=None,
        zlabel=None,
        figure=None,
        plotter=None,
        **kwords,
    ):
        """Plot a surface plot by assuming that the current dataset represents a regular matrix of points.

        Args:
            xvals (index, list or numpy.array):
                Either a column index or name or a list or numpytarray of column values. The default (None) uses
                the first column of data
            yvals (int or list):
                Either a row index or a list or numpy array of row values. The default (None) uses the column_
                headings interpreted as floats
            rectang (tuple):
                a tuple of either 2 or 4 elements representing either the origin (row,column) or size (origin,
                number of rows, number of columns) of data to be used for the z0data matrix

        Keyword Arguments:
            cmap (matplotlib colour map):
                Surface colour map - defaults to the jet colour map
            show_plot (bool):
                True Turns on interactive plot control
            title (string):
                Optional parameter that specifies the plot title - otherwise the current DataFile filename is used
            xlabel (string):
                X axes label. Default is None - guess from xvals or metadata
            ylabel (string):
                Y axes label, Default is None - guess from metadata
            zlabel (string):
                Z axis label, Default is None - guess from metadata
            figure (matplotlib figure):
                Controls what matplotlib figure to use. Can be an integer, or a matplotlib.figure or False. If
                False then a new figure is always used, otherwise it will default to using the last figure used
                by this DataFile object.
            plotter (callable):
                Optional argument that passes a plotting function into the routine. Sensible choices might be
                plt.plot (default), py.semilogy, plt.semilogx
            kwords (dict):
                A dictionary of other keyword arguments to pass into the plot function.

            Returns:
                The matplotib figure with the data plotted
        """
        # Sortout yvals values
        if isinstance(yvals, int):  # Int means we're specifying a data row
            if rectang is None:  # we need to initialise the rectang
                rectang = (yvals + 1, 0)  # We'll sort the column origin later
            elif (
                isinstance(rectang, tuple) and rectang[1] <= yvals
            ):  # We have a rectang, but we need to adjust the row origin
                rectang[0] = yvals + 1
            yvals = self[yvals]  # change the yvals into a numpy array
        elif isinstance(yvals, (list, tuple, np.ndarray)):  # We're given the yvals as a list already
            yvals = np.array(yvals)
        elif yvals is None:  # No yvals, so we'l try column headings
            if isinstance(xvals, index_types):  # Do we have an xcolumn header to take away ?
                xvals = self.find_col(xvals)
                headers = self.column_headers[xvals + 1 :]
            elif xvals is None:  # No xvals so we're going to be using the first column
                xvals = 0
                headers = self.column_headers[1:]
            else:
                headers = self.column_headers
            yvals = np.array([float(x) for x in headers])  # Ok try to construct yvals array
        else:
            raise RuntimeError("uvals must be either an integer, list, tuple, numpy array or None")
        # Sort out xvls values
        if isinstance(xvals, index_types):  # String or int means using a column index
            if xlabel is None:
                xlabel = self._col_label(xvals)
            if rectang is None:  # Do we need to init the rectan ?
                rectang = (0, xvals + 1)
            elif isinstance(rectang, tuple):  # Do we need to adjust the rectan column origin ?
                rectang[1] = xvals + 1
            xvals = self.column(xvals)
        elif isinstance(xvals, (list, tuple, np.ndarray)):  # Xvals as a data item
            xvals = np.array(xvals)
        elif isinstance(xvals, np.ndarray):
            pass
        elif xvals is None:  # xvals from column 0
            xvals = self.column(0)
            if rectang is None:  # and fix up rectang
                rectang = (0, 1)
        else:
            raise RuntimeError("xvals must be a string, integer, list, tuple or numpy array or None")

        if isinstance(rectang, tuple) and len(rectang) == 2:  # Sort the rectang value
            rectang = (
                rectang[0],
                rectang[1],
                np.shape(self.data)[0] - rectang[0],
                np.shape(self.data)[1] - rectang[1],
            )
        elif rectang is None:
            rectang = (0, 0, np.shape(self.data)[0], np.shape(self.data)[1])
        elif isinstance(rectang, tuple) and len(rectang) == 4:  # Ok, just make sure we have enough data points left.
            rectang = (
                rectang[0],
                rectang[1],
                min(rectang[2], np.shape(self.data)[0] - rectang[0]),
                min(rectang[3], np.shape(self.data)[1] - rectang[1]),
            )
        else:
            raise RuntimeError("rectang should either be a 2 or 4 tuple or None")

        # Now we can create X,Y and Z 2D arrays
        zdata = self.data[rectang[0] : rectang[0] + rectang[2], rectang[1] : rectang[1] + rectang[3]]
        xvals = xvals[0 : rectang[2]]
        yvals = yvals[0 : rectang[3]]
        xdata, ydata = np.meshgrid(xvals, yvals)

        # This is the same as for the plot_xyz routine'
        if isinstance(figure, int):
            figure, _ = self.template.new_figure(figure)
        elif isinstance(figure, bool) and not figure:
            figure, _ = self.template.new_figure(None)
        elif isinstance(figure, mplfig.Figure):
            figure, _ = self.template.new_figure(figure.number)
        elif isinstance(self.__figure, mplfig.Figure):
            figure = self.__figure
        else:
            figure, _ = self.template.new_figure(None, projection="3d")
        self.__figure = figure
        if show_plot:
            plt.ion()
        if plotter is None:
            plotter = self._surface_plotter
        plotter(xdata, ydata, zdata, cmap=cmap, **kwords)
        labels = {"xlabel": (xlabel, "X Data"), "ylabel": (ylabel, "Y Data"), "zlabel": (zlabel, "Z Data")}
        for label in labels:
            (v, default) = labels[label]
            if v is None:
                if label in self.metadata:
                    labels[label] = self[label]
                else:
                    labels[label] = default
            else:
                labels[label] = v

        plt.xlabel(str(labels["xlabel"]))
        plt.ylabel(str(labels["ylabel"]))
        if plotter is self._surface_plotter:
            self.axes[0].set_zlabel(str(labels["zlabel"]))
        if title == "":
            title = self.filename
        plt.title(title)
        plt.show()
        plt.draw()

        return self.showfig

    def plot_xy(self, xcol=None, ycol=None, fmt=None, xerr=None, yerr=None, **kargs):
        """Makesa simple X-Y plot of the specified data.

        Args:
            xcol (index):
                Xcolumn index or label
            ycol (index):
                Y column index or label

        Keyword Arguments:
            fmt (strong or sequence of strings):
                Specifies the format for the plot - see matplotlib documentation for details
            xerr,yerr (index): C
            olumns of data to get x and y errorbars from. Setting these turns the default plotter to plt.errorbar
            xlabel (string):
                X axes label. Default is None - guess from xvals or metadata
            ylabel (string):
                Y axes label, Default is None - guess from metadata
            title (string):
                Optional parameter that specifies the plot title - otherwise the current DataFile filename is used
            plotter (function):
                Function to use to plot data. Defaults to plt.plot unless error bars are set
            show_plot (bool):
                Turn on interfactive plotting and show plot when drawn
            save_filename (string or None):
                If set to a string, save the plot with this filename
            figure (integer or matplotlib.figure or boolean):
                Controls which figure is used for the plot, or if a new figure is opened.
            multiple (string):
                how to handle multiple y-axes with a common x axis. Options are:
                    -  *common* single y-axis (default)
                    -  *panels* panels sharing common x axis
                    -  *sub plots* sub plots
                    -  *y2* single axes with 2 y scales

            **kargs (dict):
                Other arguments are passed on to the plotter.

        Returns:
            A matplotlib.figure instance
        """
        c = self._fix_cols(xcol=xcol, ycol=ycol, xerr=xerr, yerr=yerr, scalar=False, **kargs)
        (kargs["xerr"], kargs["yerr"]) = (c.xerr, c.yerr)

        self.template = kargs.pop("template", self.template)
        title = kargs.pop("title", self.basename)

        defaults = {
            "capsize": 4,
            "plotter": plt.plot,
            "show_plot": True,
            "figure": self.__figure,
            "title": title,
            "save_filename": None,
            "xlabel": self._col_label(self.find_col(c.xcol)),
            "ylabel": self._col_label(self.find_col(c.ycol), True),
        }
        otherargs = []
        if "plotter" not in kargs and (
            c.xerr is not None or c.yerr is not None
        ):  # USe and errorbar blotter by default for errors
            kargs["plotter"] = plt.errorbar
            otherargs = [
                "agg_filter",
                "alpha",
                "animated",
                "antialiased",
                "aa",
                "axes",
                "clip_box",
                "clip_on",
                "clip_path",
                "color",
                "c",
                "contains",
                "dash_capstyle",
                "dash_joinstyle",
                "dashes",
                "drawstyle",
                "fillstyle",
                "gid",
                "label",
                "linestyle",
                "ls",
                "linewidth",
                "lw",
                "lod",
                "marker",
                "markeredgecolor",
                "mec",
                "markeredgewidth",
                "mew",
                "markerfacecolor",
                "mfc",
                "markerfacecoloralt",
                "mfcalt",
                "markersize",
                "ms",
                "markevery",
                "path_effects",
                "picker",
                "pickradius",
                "rasterized",
                "sketch_params",
                "snap",
                "solid_capstyle",
                "solid_joinstyle",
                "transform",
                "url",
                "visible",
                "xdata",
                "ydata",
                "zorder",
            ]
        elif "plotter" not in kargs:
            kargs["plotter"] = plt.plot
            otherargs = [
                "agg_filter",
                "alpha",
                "animated",
                "antialiased",
                "aa",
                "axes",
                "clip_box",
                "clip_on",
                "clip_path",
                "color",
                "c",
                "contains",
                "dash_capstyle",
                "dash_joinstyle",
                "dashes",
                "drawstyle",
                "fillstyle",
                "gid",
                "label",
                "linestyle",
                "ls",
                "linewidth",
                "lw",
                "lod",
                "marker",
                "markeredgecolor",
                "mec",
                "markeredgewidth",
                "mew",
                "markerfacecolor",
                "mfc",
                "markerfacecoloralt",
                "mfcalt",
                "markersize",
                "ms",
                "markevery",
                "path_effects",
                "picker",
                "pickradius",
                "rasterized",
                "sketch_params",
                "snap",
                "solid_capstyle",
                "solid_joinstyle",
                "transform",
                "url",
                "visible",
                "xdata",
                "ydata",
                "zorder",
            ]

        multiple = kargs.pop("multiple", self.multiple)

        kargs, nonkargs, fig_kargs = self._fix_kargs(None, defaults, otherargs, **kargs)

        for err in ["xerr", "yerr"]:  # Check for x and y error keywords
            if isnone(kargs.get(err, None)):
                kargs.pop(err, None)

            elif isinstance(kargs[err], index_types):
                kargs[err] = self.column(kargs[err])
            elif isiterable(kargs[err]) and isinstance(c.ycol, list) and len(kargs[err]) <= len(c.ycol):
                # Ok, so it's a list, so redo the check for each  item.
                kargs[err].extend([None] * (len(c.ycol) - len(kargs[err])))
                for i in range(len(kargs[err])):
                    if isinstance(kargs[err][i], index_types):
                        kargs[err][i] = self.column(kargs[err][i])
                    else:
                        kargs[err][i] = np.zeros(len(self))
            elif isiterable(kargs[err]) and len(kargs[err]) == len(self):
                kargs[err] = np.array(kargs[err])
            elif isinstance(kargs[err], float):
                kargs[err] = np.ones(len(self)) * kargs[err]
            else:
                kargs[err] = np.zeros(len(self))

        temp_kwords = copy.copy(kargs)
        if isinstance(c.ycol, (index_types)):
            c.ycol = [c.ycol]
        if len(c.ycol) > 1:
            if multiple == "panels":
                self.__figure, _ = plt.subplots(
                    nrows=len(c.ycol), sharex=True, gridspec_kw={"hspace": 0}, layout="constrained", **fig_kargs
                )
            elif multiple == "subplots":
                m = int(np.floor(np.sqrt(len(c.ycol))))
                n = int(np.ceil(len(c.ycol) / m))
                self.__figure, _ = plt.subplots(nrows=m, ncols=n, layout="constrained", **fig_kargs)
            else:
                self.__figure, _ = self._fix_fig(nonkargs["figure"], **fig_kargs)
        else:
            self.__figure, _ = self._fix_fig(nonkargs["figure"], **fig_kargs)
        for ix, this_yc in enumerate(c.ycol):
            if multiple != "common":
                nonkargs["ylabel"] = self._col_label(this_yc)
            if ix > 0:
                if multiple == "y2" and ix == 1:
                    self.y2()
                    lines = plt.gca()._get_lines
                    if hasattr(lines, "color_cycle"):  # mpl<1.5
                        cc = lines.color_cycle
                        next(cc)
                    elif hasattr(lines, "prop_cycler"):  # MPL<3.8.0
                        cc = lines.prop_cycler
                        next(cc)
                    else:
                        lines.get_next_color()
            if len(c.ycol) > 1 and multiple in ["y2", "panels", "subplots"]:
                self.ax = ix  # We're manipulating the plotting here
            if isinstance(fmt, list):  # Fix up the format
                fmt_t = fmt[ix]
            else:
                fmt_t = fmt
            if "label" in kargs and isinstance(kargs["label"], list):  # Fix label keywords
                temp_kwords["label"] = kargs["label"][ix]
            if "yerr" in kargs and isinstance(kargs["yerr"], list):  # Fix yerr keywords
                temp_kwords["yerr"] = kargs["yerr"][ix]
            # Call plot

            # Do interpolation of metadata
            for k in temp_kwords:
                if isinstance(temp_kwords[k], string_types) and "$" not in temp_kwords[k]:
                    temp_kwords[k] = temp_kwords[k].format(**self)
            temp_nonkargs = {}
            for k in nonkargs:
                if isinstance(nonkargs[k], string_types) and "$" not in nonkargs[k]:
                    temp_nonkargs[k] = nonkargs[k].format(**self)
                else:
                    temp_nonkargs[k] = nonkargs[k]

            self._Plot(c.xcol, c.ycol[ix], fmt_t, temp_nonkargs["plotter"], self.__figure, **temp_kwords)
            self._fix_titles(ix, multiple, **temp_nonkargs)
            if ix > 0:  # Hooks for multiple subplots
                if multiple == "panels":
                    loc, lab = plt.yticks()
                    lab = [label.get_text() for label in lab]
                    plt.yticks(loc[:-1], lab[:-1])
        return self.showfig

    def plot_xyz(self, xcol=None, ycol=None, zcol=None, shape=None, xlim=None, ylim=None, projection="3d", **kargs):
        """Plot a surface plot based on rows of X,Y,Z data using matplotlib.pcolor().

        Args:
            xcol (index):
                Xcolumn index or label
            ycol (index):
                Y column index or label
            zcol (index):
                Z column index or label

        Keyword Arguments:
            shape (tuple):
                Defines the shape of the surface (i.e. the number of X and Y value. If not provided or None, then
                the routine will attempt to calculate
                these from the data provided
            xlim (tuple):
                Defines the x-axis limits and grid of the data to be plotted
            ylim (tuple):
                Defines the Y-axis limits and grid of the data data to be plotted
            cmap (matplotlib colour map):
                Surface colour map - defaults to the jet colour map
            show_plot (bool):
                True Turns on interactive plot control
            title (string):
                Optional parameter that specifies the plot title - otherwise the current DataFile filename is used
            save_filename (string):
                Filename used to save the plot
            figure (matplotlib figure):
                Controls what matplotlib figure to use. Can be an integer, or a matplotlib.figure or False. If
                False then a new figure is always used, otherwise it will default to using the last figure used
                by this DataFile object.
            plotter (callable):
                Optional argument that passes a plotting function into the routine. Default is a 3d surface
                plotter, but contour plot and pcolormesh also work.
            projection (string or None):
                Whether to use a 3D projection or regular 2D axes (default is 3D)
            **kargs (dict):
                A dictionary of other keyword arguments to pass into the plot function.

        Returns:
            A matplotlib.figure instance
        """
        if not _3D:
            raise RuntimeError("3D plotting Not available. Install matplotlib toolkits")
        c = self._fix_cols(xcol=xcol, ycol=ycol, zcol=zcol, scalar=True, **kargs)
        if kargs.pop("griddata", True):
            xdata, ydata, zdata = self.griddata(c.xcol, c.ycol, c.zcol, shape=shape, xlim=xlim, ylim=ylim)
            cstride = int(max(1, zdata.shape[0] / 50))
            rstride = int(max(1, zdata.shape[1] / 50))
        else:
            xdata = self.column(c.xcol)
            ydata = self.column(c.ycol)
            zdata = self.column(c.zcol)
            cstride = 1
            rstride = 1

        if "template" in kargs:  # Catch template in kargs
            self.template = kargs.pop("template")

        defaults = {
            "plotter": self._surface_plotter,
            "show_plot": True,
            "figure": self.__figure,
            "title": os.path.basename(self.filename),
            "save_filename": None,
            "cmap": cm.jet,
            "rstride": cstride,
            "cstride": rstride,
        }
        coltypes = {"xlabel": c.xcol, "ylabel": c.ycol, "zlabel": c.zcol}
        for k in coltypes:
            if isinstance(coltypes[k], index_types):
                label = self._col_label(coltypes[k])
                if isinstance(label, list):
                    label = ",".join(label)
                defaults[k] = label
        if "plotter" not in kargs or ("plotter" in kargs and kargs["plotter"] is self._surface_plotter):
            otherkargs = [
                "rstride",
                "cstride",
                "color",
                "cmap",
                "facecolors",
                "norm",
                "vmin",
                "vmax",
                "shade",
                "linewidth",
                "ax",
                "alpha",
            ]
        else:
            otherkargs = ["vmin", "vmax", "shade", "color", "linewidth", "marker"]
        plotter = kargs.get("plotter", defaults["plotter"])
        self.__figure, ax = self._fix_fig(kargs.get("figure", defaults["figure"]), projection=projection)
        if isinstance(plotter, string_types):
            plotter = ax.__getattribute__(plotter)
        kargs, nonkargs, _ = self._fix_kargs(plotter, defaults, otherkargs=otherkargs, projection=projection, **kargs)
        self.plot3d = plotter(xdata, ydata, zdata, **kargs)
        if plotter is not self._surface_plotter:
            del nonkargs["zlabel"]
        self._fix_titles(0, "none", **nonkargs)
        return self.showfig

    def plot_xyuv(self, xcol=None, ycol=None, ucol=None, vcol=None, wcol=None, **kargs):
        """Make an overlaid image and quiver plot.

          Args:
        !c      xcol (index):
                  Xcolumn index or label
              ycol (index):
                  Y column index or label
              zcol (index):
                  Z column index or label
              ucol (index):
                  U column index or label
              vcol (index):
                  V column index or label
              wcol (index):
                  W column index or label

          Keyword Arguments:
              show_plot (bool):
                  True Turns on interactive plot control
              title (string):
                  Optional parameter that specifies the plot title - otherwise the current DataFile filename is used
              save_filename (string):
                  Filename used to save the plot
              figure (matplotlib figure):
                  Controls what matplotlib figure to use. Can be an integer, or a matplotlib.figure or False. If False
                  then a new figure is always used, otherwise it will default to using the last figure used by this
                  DataFile object.
              no_quiver (bool):
                  Do not overlay quiver plot (in cases of dense meshes of points)
              plotter (callable):
                  Optional argument that passes a plotting function into the routine. Default is a 3d surface plotter,
                  but contour plot and pcolormesh also work.
              **kargs (dict):
                  A dictionary of other keyword arguments to pass into the plot function.
        """
        c = self._fix_cols(xcol=xcol, ycol=ycol, ucol=ucol, vcol=vcol, wcol=wcol, **kargs)
        Z = self._vector_color(xcol=xcol, ycol=ycol, ucol=ucol, vcol=vcol, wcol=wcol)
        if "template" in kargs:  # Catch template in kargs
            self.template = kargs.pop("template")
        no_quiver = kargs.pop("no_quiver", False)

        if "save_filename" in kargs:
            save = kargs["save_filename"]
            del kargs["save_filename"]
        else:
            save = None
        kargs.setdefault("alpha", 0.75)
        fig = self.image_plot(c.xcol, c.ycol, Z, **kargs)
        if save is not None:  # stop saving file twice
            kargs["save_filename"] = save
        if not no_quiver:
            fig = self.quiver_plot(c.xcol, c.ycol, c.ucol, c.vcol, **kargs)

        return fig

    plot_xyuvw = plot_xyuv

    def plot_xyzuvw(self, xcol=None, ycol=None, zcol=None, ucol=None, vcol=None, wcol=None, **kargs):
        """Plot a vector field plot based on rows of X,Y,Z (U,V,W) data using ,ayavi.

        Args:
            xcol (index):
                Xcolumn index or label
            ycol (index):
                Y column index or label
            zcol (index):
                Z column index or label
            ucol (index):
                U column index or label
            vcol (index):
                V column i
                ndex or label
            wcol (index): W column index or label

        Keyword Arguments:
            colormap (string):
                Vector field colour map - defaults to the jet colour map
            colors (column index or numpy array):
                Values used to map the colors of the resultant file.
            mode (string):
                glyph type, default is "cone"
            scale_factor(float):
                Scale-size of glyphs.
            figure (mlab figure):
                Controls what mlab figure to use. Can be an integer, or a mlab.figure or False. If False then a new
                figure is always used,
                otherwise it will default to using the last figure used by this DataFile object.
            plotter (callable):
                Optional argument that passes a plotting function into the routine. Sensible choices might be
                plt.plot (default), py.semilogy, plt.semilogx
            kargs (dict):
                A dictionary of other keyword arguments to pass into the plot function.

        Returns:
            A mayavi scene instance
        """
        try:
            from mayavi import mlab, core

            mlab.figure()
            mayavi = True
        except ImportError:
            mayavi = False
        c = self._fix_cols(xcol=xcol, ycol=ycol, zcol=zcol, ucol=ucol, vcol=vcol, wcol=wcol, scalar=True, **kargs)

        if "template" in kargs:  # Catch template in kargs
            self.template = kargs.pop("template")

        if mayavi:
            defaults = {
                "figure": self.__figure,
                "plotter": self._vector_field_plotter,
                "show_plot": True,
                "mode": "cone",
                "scale_factor": 1.0,
                "colors": True,
            }
            otherkargs = [
                "color",
                "colormap",
                "extent",
                "figure",
                "line_width",
                "mask_points",
                "mode",
                "name",
                "opacity",
                "reset_zoom",
                "resolution",
                "scalars",
                "scale_factor",
                "scale_mode",
                "transparent",
                "vmax",
                "vmin",
            ]
        else:
            defaults = {
                "plotter": globals()["__mpl3DQuiver"],
                "show_plot": True,
                "figure": self.__figure,
                "title": os.path.basename(self.filename),
                "save_filename": None,
                "cmap": cm.jet,
                "scale": 1.0,
                "units": "xy",
                "color": hsl2rgb((1 + self.q / np.pi) / 2, self.r / np.max(self.r), (1 + self.w) / 2, alpha=True)
                / 255.0,
            }
            projection = kargs.pop("projection", "3d")
            coltypes = {"xlabel": c.xcol, "ylabel": c.ycol, "zlabel": c.zcol}
            for k in coltypes:
                if isinstance(coltypes[k], index_types):
                    label = self._col_label(coltypes[k])
                    if isinstance(label, list):
                        label = ",".join(label)
                    defaults[k] = label
            if "plotter" not in kargs or ("plotter" in kargs and kargs["plotter"] is __mpl3DQuiver):
                otherkargs = ["color", "cmap", "linewidth", "ax", "length", "pivot", "arrow_length_ratio"]
            else:
                otherkargs = ["color", "linewidth"]

        kargs, nonkargs, _ = self._fix_kargs(None, defaults, otherkargs=otherkargs, **kargs)
        colors = nonkargs.pop("color", True)
        if isinstance(colors, bool) and colors:
            pass
        elif isinstance(colors, index_types):
            colors = self.column(colors)
        elif isinstance(colors, np.ndarray):
            pass
        elif callable(colors):
            colors = np.array([colors(x) for x in self.rows()])
        else:
            raise RuntimeError("Do not recognise what to do with the colors keyword.")
        if mayavi:
            kargs["scalars"] = colors
        figure = nonkargs["figure"]
        plotter = nonkargs["plotter"]
        if mayavi:
            if isinstance(figure, int):
                figure = mlab.figure(figure)
            elif isinstance(figure, bool) and not figure:
                figure = mlab.figure(bgcolor=(1, 1, 1))
            elif isinstance(figure, core.scene.Scene):
                pass
            elif isinstance(self.__figure, core.scene.Scene):
                figure = self.__figure
            else:
                figure = mlab.figure(bgcolor=(1, 1, 1))
            self.__figure = figure

        else:
            self.__figure, ax = self._fix_fig(nonkargs["figure"], projection=projection)
            if isinstance(plotter, string_types):
                plotter = ax.__getattribute__(plotter)

        kargs["figure"] = figure
        plotter(
            self.column(c.xcol),
            self.column(c.ycol),
            self.column(c.zcol),
            self.column(c.ucol),
            self.column(c.vcol),
            self.column(c.wcol),
            **kargs,
        )
        if nonkargs["show_plot"]:
            if mayavi:
                mlab.show()
            else:
                plt.show()
        return self.showfig

    def plot_voxels(self, xcol=None, ycol=None, zcol=None, ucol=None, cmap=None, **kargs):
        """Make a volumetric plot of data arranged as x,y,z,u.

        Args:
            xcol (index):
                Xcolumn index or label
            ycol (index):
                Y column index or label
            zcol (index):
                Z column index or label
            ucol (index):
                U column index or label

        Keyword Arguments:
            visible (callable):
                A function f(x,y,z) that returns True if a voxcel is to be visible
            cmap (colourmap):
                A Matplotlib colour map to apply for the magnitude (u column) data.

        Returns:
            (matplotlib.Figure):
                The figure window contacting the plot

        Example:
            .. plot:: samples/voxel_plot.py
                :include-source:
                :outname: voxels


        """
        _ = self._fix_cols(xcol=xcol, ycol=ycol, zcol=zcol, ucol=ucol, **kargs)
        defaults = {
            "plotter": self.plot_voxels,
            "show_plot": True,
            "figure": self.__figure,
            "title": os.path.basename(self.filename),
            "xlabel": self._col_label(self.find_col(_.xcol)),
            "ylabel": self._col_label(self.find_col(_.ycol)),
            "zlabel": self._col_label(self.find_col(_.zcol)),
            "save_filename": None,
            "cmap": cm.viridis,
            "f_alpha": 0.5,
            "e_alpha": 0.9,
            "filled": None,
        }
        otherkargs = {}
        shape = kargs.pop("shape", None)
        xlim = kargs.pop("xlim", None)
        ylim = kargs.pop("ylim", None)
        zlim = kargs.pop("zlim", None)
        X, Y, Z, U = self.griddata(_.xcol, _.ycol, _.zcol, _.ucol, shape=shape, xlim=xlim, ylim=ylim, zlim=zlim)

        if callable(kargs.get("visible", False)):
            visible = kargs.pop("visible")
            try:
                filled = visible(self // _, xcol, self // _.ycol, self // _.zcol)
            except (ValueError, TypeError, RuntimeError):
                filled = np.array([visible(*pt) for pt in zip(self // _.xcol, self // _.ycol, self // _.zcol)])
            filled = np.where(filled, np.ones_like(filled), np.zeros_like(filled))
            filled = (
                self.griddata(_.xcol, _.ycol, _.zcol, filled, shape=shape, xlim=xlim, ylim=ylim, zlim=zlim)[3] >= 0.5
            )
        else:
            filled = kargs.pop("filled", np.ones_like(U, dtype=bool))

        if "template" in kargs:  # Catch template in kargs
            self.template = kargs.pop("template")
        kargs, nonkargs, _ = self._fix_kargs(None, defaults, otherkargs=otherkargs, **kargs)
        self.__figure, ax = self._fix_fig(nonkargs["figure"], projection="3d")

        norm = colors.Normalize(vmin=U.min(), vmax=U.max(), clip=True)
        mapper = cm.ScalarMappable(norm=norm, cmap=kargs.get("cmap", cmap))
        cshape = U.shape + (4,)
        facecolors = mapper.to_rgba(U.ravel(), alpha=nonkargs["f_alpha"]).reshape(cshape)
        edgecolors = mapper.to_rgba(U.ravel(), alpha=nonkargs["e_alpha"]).reshape(cshape)

        ax.voxels(X, Y, Z, filled=filled, edgecolors=edgecolors, facecolors=facecolors)

        self._fix_titles(0, "none", **nonkargs)
        return self.showfig

    def quiver_plot(self, xcol=None, ycol=None, ucol=None, vcol=None, **kargs):
        """Make a 2D Quiver plot from the data.

        Args:
            xcol (index):
                Xcolumn index or label
            ycol (index):
                Y column index or label
            zcol (index):
                Z column index or label
            ucol (index):
                U column index or label
            vcol (index):
                V column i
                ndex or label
            wcol (index): W column index or label

        Keyword Arguments:
            xlabel (string):
                X axes label. Default is None - guess from xvals or metadata
            ylabel (string):
                Y axes label, Default is None - guess from metadata
            zlabel (string):
                Z axis label, Default is None - guess from metadata
            plotter (function):
                Function to use to plot data. Defaults to plt.contour
            headlength,headwidth,headaxislength (float):
                Controls the size of the quiver heads
            show_plot (bool):
                Turn on interfactive plotting and show plot when drawn
            save_filename (string or None):
                If set to a string, save the plot with this filename
            figure (integer or matplotlib.figure or boolean):
                Controls which figure is used for the plot, or if a new figure is opened.
            **kargs (dict):
                Other arguments are passed on to the plotter.


        Returns:
            A matplotlib figure instance.

        Keyword arguments are all passed through to :py:func:`matplotlib.plt.quiver`.

        """
        locals().update(self._fix_cols(xcol=xcol, ycol=ycol, ucol=ucol, vcol=vcol, **kargs))
        defaults = {
            "pivot": "mid",
            "color": (0, 0, 0),
            "headlength": 5,
            "headaxislength": 5,
            "headwidth": 4,
            "units": "xy",
            "plotter": plt.quiver,
            "show_plot": True,
            "figure": self.__figure,
            "title": os.path.basename(self.filename),
            "xlabel": self._col_label(self.find_col(xcol)),
            "ylabel": self._col_label(self.find_col(ycol)),
        }
        otherkargs = [
            "units",
            "angles",
            "scale",
            "scale_units",
            "width",
            "headwidth",
            "headlength",
            "headaxislength",
            "minshaft",
            "minlength",
            "pivot",
            "color",
        ]

        if "template" in kargs:  # Catch template in kargs
            self.template = kargs.pop("template")

        kargs, nonkargs, _ = self._fix_kargs(None, defaults, otherkargs=otherkargs, **kargs)
        plotter = nonkargs["plotter"]
        self.__figure, _ = self._fix_fig(nonkargs["figure"])
        data = np.column_stack([self // xcol, self // ycol, self // ucol, self // vcol])

        fig = plotter(
            data[:, 0],
            data[:, 1],
            data[:, 2],
            data[:, 3],
            **kargs,
        )
        self._fix_titles(0, "non", **nonkargs)
        return fig

    def subplot(self, *args, **kargs):
        """Pass throuygh for :py:func:`matplotlib.pyplot.subplot`.

        Args:
            rows (int):
                If this is the only argument, then a three digit number representing
                the rows,columns,index arguments. If separate rows, column and index are provided,
                then this is the number of rows of sub-plots in one figure.
            columns (int):
                The number of columns of sub-plots in one figure.
            index (int):
                Index (1 based) of the current sub-plot.

        Returns:
            A matplotlib.Axes instance representing the current sub-plot

        As well as passing through to the plyplot routine of the same name, this
        function maintains a list of the current sub-plot axes via the subplots attribute.
        """
        self.template.new_figure(self.__figure.number, no_axes=True)
        sp = plt.subplot(*args, **kargs)
        if len(args) == 1:
            rows = args[0] // 100
            cols = (args[0] // 10) % 10
            index = args[0] % 10
        else:
            rows = args[0]
            cols = args[1]
            index = args[2]
        if len(self._subplots) < rows * cols:
            self._subplots.extend([None for i in range(rows * cols - len(self._subplots))])
        self._subplots[index - 1] = sp
        return sp

    def subplot2grid(self, *args, **kargs):
        """Provide a pass through to :py:func:`matplotlib.pyplot.subplot2grid`."""
        if self.__figure is None:
            self.figure(no_axes=True)

        figure = self.template.new_figure(self.__figure.number, no_axes=True)[0]

        plt.figure(figure.number)
        ret = plt.subplot2grid(*args, **kargs)
        return ret

    def x2(self):
        """Generate a new set of axes with a second x-scale.

        Returns:
            The new matplotlib.axes instance.
        """
        ax = self.fig.gca()
        ax2 = ax.twiny()
        plt.sca(ax2)
        return ax2

    def y2(self):
        """Generate a new set of axes with a second y-scale.

        Returns:
            The new matplotlib.axes instance
        """
        ax = self.fig.gca()
        ax2 = ax.twinx()
        # plt.subplots_adjust(right=self.__figure.subplotpars.right - 0.05)
        plt.sca(ax2)
        return ax2



# -*- coding: utf-8 -*-
"""
Created on Sat Dec  8 20:59:16 2018

@author: phygbu
"""

import sys
import os.path as path
import os
import pytest

from Stoner import DataFolder
from Stoner.Util import hysteresis_correct

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../../"))
sys.path.insert(0, pth)
datadir = path.join(pth, "sample-data")


def test_each_call():
    os.chdir(datadir)
    fldr6 = DataFolder(".", pattern="QD*.dat", pruned=True)
    shaper = lambda f: f.shape
    fldr6.sort()
    res = fldr6.each(shaper)
    assert res == [(6049, 88), (3026, 41), (1410, 57), (412, 72)], "__call__ on each fauiled."
    fldr6.each.del_column(0)
    res = fldr6.each(shaper)
    assert res == [(6049, 87), (3026, 40), (1410, 56), (412, 71)], "Proxy method call via each failed"
    paths = ["QD-MH.dat", "QD-PPMS.dat", "QD-PPMS2.dat", "QD-SQUID-VSM.dat"]
    filenames = [path.relpath(x, start=fldr6.directory) for x in fldr6.each.filename.tolist()]
    assert filenames == paths, "Reading attributes from each failed."
    meths = [x for x in dir(fldr6.each) if not x.startswith("_")]
    assert len(meths) == 138, "Dir of folders.each failed ({}).".format(len(meths))


def test_each_call_or_operator():
    os.chdir(datadir)
    fldr4 = DataFolder(datadir, pattern="QD-SQUID-VSM.dat")
    fldr5 = fldr4.clone
    (hysteresis_correct @ fldr4)(setas="3.xy", saturated_fraction=0.25)
    assert "Hc" in fldr4[0], "Matrix multiplication of callable by DataFolder failed test."
    fldr5.each(hysteresis_correct, setas="3.xy", saturated_fraction=0.25)
    assert "Hc" in fldr5[0], "Call on DataFolder.each() failed to apply function to folder"


def test_each_setas():
    os.chdir(datadir)
    fldr7 = DataFolder("NLIV", pattern="*.txt", pruned=True, setas="yx.")
    assert len(fldr7.each.setas) == 3, "Length of DataFolder.each.setas wrong"
    assert fldr7.each.setas.collapse() == ["y", "x", "."], "DataFolder collapsed each.setas wrong"
    del fldr7.each.setas[1]
    assert fldr7.each.setas.collapse() == ["y", ".", "."], "DataFolder.each.setas delitem failed"
    fldr7.each.setas[1] = "z"
    assert fldr7.each.setas[1] == ["z"] * len(fldr7), "DataFolder.each.setas setitem and getitem failed"
    fldr7.each.setas = "yx."
    assert fldr7.each.setas.collapse() == ["y", "x", "."], "DataFolder.each.setas assignment failed"


def test_each_attr():
    os.chdir(datadir)
    fldr6 = DataFolder(".", pattern="QD*.dat", pruned=True)
    with pytest.raises(AttributeError):
        _ = fldr6.each.bad_item
    fldr6.each.column_headers = ["X", "Y"]
    assert fldr6.each.column_headers.size == 4, "Read back of an attribute failed to return the right array length"
    res = [x[0] for x in fldr6.each.column_headers]
    assert res == ["X"] * 4, "Setting DataFolder.each.attr failed"


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



"""Test langevin fitting."""

# pylint: disable=invalid-name
from copy import copy

from numpy import linspace, ones_like
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.generic import (
    Lorentzian_diff,
    lorentzian_diff,
)

x = linspace(-1.0, 1.0, 101)
params = [1, 0.1, -0.25]
y = lorentzian_diff(x, *params) + normal(size=len(x), scale=0.5)
dy = ones_like(x) * 5e-3
d = Data(x, y, dy, setas="xye", column_headers=["Time", "Signal", "dM"])

d.curve_fit(lorentzian_diff, p0=copy(params), result=True, header="curve_fit")

d.setas = "xye"

d.lmfit(Lorentzian_diff, result=True, header="lmfit", prefix="lmfit")


d.setas = "xyeyy"
d.plot(fmt=["r+", "b-", "g-"])

d.annotate_fit(
    lorentzian_diff,
    x=0.6,
    y=0.2,
    fontdict={"size": "x-small", "color": "blue"},
    mode="eng",
)
d.annotate_fit(
    Lorentzian_diff,
    x=0.05,
    y=0.2,
    fontdict={"size": "x-small", "color": "green"},
    prefix="lmfit",
    mode="eng",
)
d.title = "Differential Lorentzian Fit"



"""Test langevin fitting."""

# pylint: disable=invalid-name
from numpy import linspace, ones_like, where, nan, isnan
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.magnetism import blochLaw, BlochLaw

T = linspace(10, 1000, 100)
Ms = 1714
Tc = 1046
M = where(
    T > 300, nan, blochLaw(T, Ms, Tc) + normal(scale=Ms / 200, size=len(T))
)
dM = ones_like(T) * Ms / 200

d = Data(
    T, M, dM, column_headers=["Temperature", "Magnetization", "dM"], setas="xye"
)

d.curve_fit(
    blochLaw,
    p0=[1500, 1500],
    result=True,
    header="curve_fit",
    prefix="curve_fit",
    bounds=lambda x, r: not isnan(r.y),
)

model = BlochLaw()

fit = d.lmfit(
    model,
    result=True,
    header="lmfit",
    prefix="lmfit",
    output="report",
    bounds=lambda x, r: not isnan(r.y),
    g=2.01,
)


d.setas = "xyeyy"
d.plot(fmt=["r.", "b-", "g-"])

d.annotate_fit(
    blochLaw,
    x=0.1,
    y=0.5,
    fontdict={"size": "x-small", "color": "blue"},
    mode="eng",
    prefix="curve_fit",
)
d.annotate_fit(
    BlochLaw,
    x=0.1,
    y=0.25,
    fontdict={"size": "x-small", "color": "green"},
    prefix="lmfit",
    mode="eng",
)
d.title = "BlochLaw Fit"



# -*- coding: utf-8 -*-
"""
Created on Fri May 27 17:09:04 2016

@author: phyrct
"""

from Stoner.Image.kerr import KerrArray, KerrImageFile, KerrStack
from Stoner import Data, __home__
import numpy as np
import pytest
import os
import matplotlib.pyplot as plt


import Stoner

Stoner.Options.multiprocessing = False

# data arrays for testing - some useful small images for tests

testdir = os.path.join(os.path.dirname(__file__), "kerr_testdata")
testdir2 = os.path.join(os.path.dirname(__file__), "coretestdata", "testims")
sample_data_dir = os.path.join(__home__, "../sample-data")


def shares_memory(arr1, arr2):
    """Check if two numpy arrays share memory"""
    # ret = arr1.base is arr2 or arr2.base is arr1
    ret = np.may_share_memory(arr1, arr2)
    return ret


selfimage = KerrArray(os.path.join(testdir, "kermit3.png"), ocr_metadata=True)
selfimage2 = KerrArray(os.path.join(testdir, "kermit3.png"))
selfimage3 = KerrImageFile(os.path.join(sample_data_dir, "testnormalsave.png"))
selfks = KerrStack(testdir2)


def test_kerr_ops():
    im = selfimage3.clone
    assert isinstance(im.image, KerrArray), "KerrImageFile not blessing the image property correctly"
    im1 = im.float_and_croptext()
    assert isinstance(im1.image, KerrArray), "Calling a crop routine without the _ argument returns a new KerrArray"
    im2 = im.float_and_croptext(_=True)
    assert im2 == im, "Calling crop method with _ argument changed the KerrImageFile"
    im = KerrImageFile(selfimage2.clone)
    im.float_and_croptext(_=True)
    mask = im.image.defect_mask_subtract_image()
    im.image[~mask] = np.mean(im.image[mask])
    _ = im
    assert mask.sum() == 343228, "Mask didn't work out right"
    im = KerrImageFile(selfimage2.clone)
    im.float_and_croptext(_=True)
    mask = im.image.defect_mask(radius=4)
    im.image[~mask] = np.mean(im.image[mask])
    selim2 = im
    assert mask.sum() == 342540, "Mask didn't work out right"
    selim2.level_image()
    selim2.remove_outliers()
    selim2.normalise()
    selim2.plot_histogram()
    selim2.imshow()
    assert len(plt.get_fignums()) == 2, "Didn't open the correct number of figures"
    plt.close("all")


def test_tesseract_ocr():
    # this incidentally tests get_metadata too
    if not selfimage.tesseractable:
        print("#" * 80)
        print("Skipping test that uses tesseract.")
        return None
    _ = selfimage.metadata

    # assert all((m['ocr_scalebar_length_microns']==50.0,
    #                     m['ocr_date']=='11/30/15',
    #                     m['ocr_field'] == -0.13)), 'Misread metadata {}'.format(m))
    _ = (
        "ocr_scalebar_length_pixels",
        "ocr_field_of_view_microns",
        "Loaded from",
        "ocr_microns_per_pixel",
        "ocr_pixels_per_micron",
    )
    # assert all([k in m.keys() for k in keys]), 'some part of the metadata didn\'t load {}'.format(m))
    m_un = selfimage2.metadata
    assert "ocr_field" not in m_un.keys(), "Unannotated image has wrong metadata"


def test_kerrstack():
    print("X" * 80 + "\n" + "Test Kerrstack")
    ks = selfks.clone
    ks.each.normalise(scale=(0, 1.0))
    assert np.min(ks.imarray) == 0.0 and np.max(ks.imarray) == 1.0, "KerrStack subtract failed min,max: {},{}".format(
        np.min(ks.imarray), np.max(ks.imarray)
    )
    d = ks.hysteresis()
    assert isinstance(d, Data), "hysteresis didn't return Data"
    assert d.data.shape == (len(ks), 2), "hysteresis didn't return correct shape"


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Loader for zip files."""
import zipfile as zf
import os.path as path
from traceback import format_exc

from ...compat import str2bytes, path_types
from ...core.exceptions import StonerLoadError
from ...tools import copy_into, make_Data
from ..decorators import register_loader, register_saver
from ..utils.zip import test_is_zip


@register_loader(patterns=(".zip", 16), mime_types=("application/zip", 16), name="ZippedFile", what="Data")
def load_zipfile(new_data, filename=None, *args, **kargs):
    """Load a file from the zip file, opening it as necessary."""
    new_data.filename = filename
    try:
        if isinstance(new_data.filename, zf.ZipFile):  # Loading from an ZipFile
            if not new_data.filename.fp:  # Open zipfile if necessary
                other = zf.ZipFile(new_data.filename.filename, "r")
                close_me = True
            else:  # Zip file is already open
                other = new_data.filename
                close_me = False
            member = kargs.get("member", other.namelist()[0])
            solo_file = len(other.namelist()) == 1
        elif isinstance(new_data.filename, path_types) and zf.is_zipfile(
            new_data.filename
        ):  # filename is a string that is a zip file
            other = zf.ZipFile(new_data.filename, "a")
            member = kargs.get("member", other.namelist()[0])
            close_me = True
            solo_file = len(other.namelist()) == 1
        else:
            raise StonerLoadError(f"{new_data.filename} does  not appear to be a real zip file")
    except StonerLoadError:
        raise
    except Exception as err:  # pylint: disable=W0703 # Catching everything else here
        try:
            exc = format_exc()
            other.close()
        except (AttributeError, NameError, ValueError, TypeError, zf.BadZipFile, zf.LargeZipFile):
            pass
        raise StonerLoadError(f"{new_data.filename} threw an error when opening\n{exc}") from err
    # Ok we can try reading now
    info = other.getinfo(member)
    data = other.read(info)  # In Python 3 this would be a bytes
    tmp = make_Data() << data.decode("utf-8")
    copy_into(tmp, new_data)
    # new_data.__init__(tmp << data)
    new_data.filename = path.join(other.filename, member)
    if close_me:
        other.close()
    if solo_file:
        new_data.filename = str(filename)
    return new_data


@register_saver(patterns=(".zip", 16), name="ZippedFile", what="Data")
def save(save_data, filename=None, **kargs):
    """Override the save method to allow ZippedFile to be written out to disc (as a mininmalist output).

    Args:
        filename (string or zipfile.ZipFile instance):
            Filename to save as (using the same rules as for the load routines)

    Returns:
        A copy of itsave_data.
    """
    if filename is None:
        filename = save_data.filename
    if filename is None or (isinstance(filename, bool) and not filename):  # now go and ask for one
        filename = save_data.__file_dialog("w")
    compression = kargs.pop("compression", zf.ZIP_DEFLATED)
    try:
        if isinstance(filename, path_types):  # We;ve got a string filename
            if test_is_zip(filename):  # We can find an existing zip file somewhere in the filename
                zipfile, member = test_is_zip(filename)
                zipfile = zf.ZipFile(zipfile, "a")
                close_me = True
            elif path.exists(filename):  # The fiule exists but isn't a zip file
                raise IOError(f"{filename} Should either be a zip file or a new zip file")
            else:  # Path doesn't exist, use extension of file part to find where the zip file should be
                parts = path.split(filename)
                for i, part in enumerate(parts):
                    if path.splitext(part)[1].lower() == ".zip":
                        break
                else:
                    raise IOError(f"Can't figure out where the zip file is in {filename}")
                zipfile = zf.ZipFile(path.join(*parts[: i + 1]), "w", compression, True)
                close_me = True
                member = path.join("/", *parts[i + 1 :])
        elif isinstance(filename, zf.ZipFile):  # Handle\ zipfile instance, opening if necessary
            if not filename.fp:
                filename = zf.ZipFile(filename.filename, "a")
                close_me = True
            else:
                close_me = False
            zipfile = filename
            member = ""

        if member == "" or member == "/":  # Is our file object a bare zip file - if so create a default member name
            if len(zipfile.namelist()) > 0:
                member = zipfile.namelist()[-1]
                save_data.filename = path.join(filename, member)
            else:
                member = "DataFile.txt"
                save_data.filename = filename

        zipfile.writestr(member, str2bytes(str(save_data)))
        if close_me:
            zipfile.close()
    except (zipfile.BadZipFile, IOError, TypeError, ValueError) as err:
        error = format_exc()
        try:
            zipfile.close()
        finally:
            raise IOError(f"Error saving zipfile\n{error}") from err
    return save_data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Defines type hint types for various common types in the package."""
# pylint: disable-all
import typing
from pathlib import PurePath
from numpy import ndarray

from ..compat import _pattern_type

# Types used in compat code
RegExp = _pattern_type
String_Types = typing.Union[str, bytes]
Int_Types = int
Path_Types = typing.Union[String_Types, PurePath]

# Column Indices
Single_Column_Index = typing.Union[str, bytes, int, RegExp]
Column_Index = typing.Union[typing.Sequence[Single_Column_Index], Single_Column_Index]

# Setas
Setas_Base = typing.Sequence[str]
Setas_Dict = typing.Mapping[str, Column_Index]
Setas = typing.Union[Setas_Base, Setas_Dict, str, "setas"]

# Other useful types
Filename = typing.Union[Path_Types, bool]
Numeric = typing.Union[float, int, complex]
NumericArray = typing.Union[Numeric, ndarray]



# -*- coding: utf-8 -*-
"""Stiner.Image.folders test
"""

from Stoner.Image import ImageArray, ImageFolder
import pytest
import os

knownkeys = [
    "Averaging",
    "Comment:",
    "Contrast Shift",
    "HorizontalFieldOfView",
    "Lens",
    "Loaded from",
    "Magnification",
    "MicronsPerPixel",
    "field: units",
    "field: units",
    "filename",
    "subtraction",
]
knownfieldvals = [
    -233.432852,
    -238.486666,
    -243.342465,
    -248.446173,
    -253.297813,
    -258.332918,
    -263.340476,
    -268.20511,
]

testdir = os.path.join(os.path.dirname(__file__), "coretestdata", "testims")
fldr = ImageFolder(testdir, pattern="*.png")


def test_load():
    fldr = ImageFolder(testdir, pattern="*.png")
    assert len(fldr) == len(os.listdir(testdir)), "Didn't find all the images"
    stack = fldr.as_stack()
    assert len(stack) == len(fldr), "Conversion to ImageStack with as_stack() failed to preserve length"


def test_properties():
    fldr = ImageFolder(testdir, pattern="*.png")
    assert fldr.size == (512, 672), f"fldr.size incorrect at {fldr.size}"
    empty = ImageFolder()
    assert empty.size == tuple(), f"Empty folder.size wrong at {empty.size}"
    fldr[1] = fldr[1].crop(4)
    assert not fldr.size, f"fldr.size didn't return False for non-uniform images, got {fldr.size}"
    for im in fldr.images:
        assert isinstance(im, ImageArray), "fldr.images returned something that wasn't an image array"


def test_methods():
    fldr = ImageFolder(testdir, pattern="*.png")
    assert len(list(fldr.loaded)) == 0, "ImageFolder got loaded unexpectedly!"
    fldr.loadgroup()
    assert len(list(fldr.loaded)) == 8, "ImageFolder.looadgroup() failed to load!"


if __name__ == "__main__":
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Script to produce plots from GenX output."""
# pylint: disable=invalid-name
import numpy as np
import matplotlib.pyplot as plt
from Stoner import DataFolder, Data


f = DataFolder(directory=False, pattern="*.dat", type=Data)
f.sort("dataset")
up = f[2]
up &= f[1]
asym = f[0]

up.column_headers = [
    "q_up",
    "Isim_up",
    "I_up",
    "dI_up",
    "q_down",
    "Isim_down",
    "I_down",
    "dI_down",
]

up.plot_xy("q_up", "Isim_up", "r-", plotter=plt.semilogy, label=None)
up.plot_xy("q_down", "Isim_down", "b-", label=None)
up.plot_xy("q_up", "I_up", "ro", yerr="dI_up", label="Up", ms=2.0)
up.plot_xy("q_down", "Isim_down", "bx", yerr="dI_down", label="Down", ms=2.0)

plt.legend()
plt.xlabel("Momentum Transfer Q($\AA^{-1}$)")
plt.ylabel("Normalised Intensity")
plt.title("Spin Up and Down Signals")

up.diffsum("I_up", "I_down", header="SpinAsym_data")
up.diffsum("Isim_up", "Isim_down", header="SpinAsym_sim")
up_c = up.find_col("I_up")
down_c = up.find_col("I_down")
up_ec = up.find_col("dI_up")
down_ec = up.find_col("dI_down")
func = lambda r: np.sqrt(
    (
        1.0 / (r[up_c] + r[down_c])
        - (r[up_c] - r[down_c]) / (r[up_c] + r[down_c]) ** 2
    )
    ** 2
    * r[up_ec] ** 2
    + (
        -1.0 / (r[up_c] + r[down_c])
        - (r[up_c] - r[down_c]) / (r[up_c] + r[down_c]) ** 2
    )
    ** 2
    * r[down_ec] ** 2
)
up.apply(func, up.find_col("SpinAsym_data"), replace=False, header="dSpinAsym")
up.fig = plt.figure()

up.plot_xy(
    "q_up",
    "SpinAsym_data",
    fmt="ro",
    yerr="dSpinAsym",
    label="Spin Asymmetry Data",
    ms=2.0,
)
up.plot_xy("q_up", "SpinAsym_sim", "r-", label="Calculated Spin Asymmetry")
plt.legend()
plt.xlabel("Momentum Transfer Q($\AA^{-1}$)")
plt.ylabel("Spin Asymmetry")
plt.title("Spin Asymmetry Data and Calculated")

asym.add("real sld +", "real sld -", header="Structure")
asym.subtract("real sld +", "real sld -", header="Magnetic")
asym.plot_xy(0, "Structure")
asym.plot_xy(0, "Magnetic")
plt.legend()
plt.xlabel("Distance z\ (nm)")
plt.ylabel("Scattering Potential")
plt.title("Magnetic and Structural Scattering Potentials")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implements DataFile like classes for various large scale facilities."""

# Standard Library imports
from ...core.exceptions import StonerLoadError

from ..decorators import register_loader

try:
    import fabio
except ImportError:
    fabio = None

if fabio:

    @register_loader(
        patterns=(".edf", 32),
        mime_types=[("text/plain", 32), ("application/octet-stream", 32)],
        name="FabioImage",
        what="Image",
    )
    def load_fabio(new_data, filename=None, *args, **kargs):
        """Load function. File format has space delimited columns from row 3 onwards."""
        if filename is None or not filename:
            new_data.get_filename("r")
        else:
            new_data.filename = filename
        try:
            img = fabio.open(new_data.filename)
            new_data.image = img.data
            new_data.metadata.update(img.header)
            return new_data
        except (OSError, ValueError, TypeError, IndexError) as err:
            raise StonerLoadError("Not a Fabio Image file !") from err



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implement DataFile classes for some generic file formats."""
import csv
import contextlib
import copy
import io
from pathlib import Path
import re
from collections.abc import Mapping
import sys
import logging
from typing import Optional
import warnings

import PIL
import numpy as np

from ...compat import str2bytes, Hyperspy_ok, hs, hsload
from ...core.array import DataArray
from ...core.exceptions import StonerLoadError
from ...core.utils import tab_delimited
from ...tools import make_Data

from ...tools.file import FileManager

from ..decorators import register_loader, register_saver


class _refuse_log(logging.Filter):
    """Refuse to log all records."""

    def filter(self, record):
        """Do not log anything."""
        return False


@contextlib.contextmanager
def catch_sysout(*args):
    """Temporarily redirect sys.stdout and.sys.stdin."""
    stdout, stderr = sys.stdout, sys.stderr
    out = io.StringIO()
    sys.stdout, sys.stderr = out, out
    logger = logging.getLogger("hyperspy.io")
    logger.addFilter(_refuse_log)
    yield None
    logger.removeFilter(_refuse_log)
    sys.stdout, sys.stderr = stdout, stderr
    return


def _delim_detect(line):
    """Detect a delimiter in a line.

    Args:
        line(str):
            String to search for delimiters in.

    Returns:
        (str):
            Delimiter to use.

    Raises:
        StnerLoadError:
            If delimiter cannot be located.
    """
    quotes = re.compile(r"([\"\'])[^\1]*\1")
    line = quotes.sub("", line)  # Remove quoted strings first
    current = (None, len(line))
    for delim in "\t ,;":
        try:
            idx = line.index(delim)
        except ValueError:
            continue
        if idx < current[1]:
            current = (delim, idx)
    if current[0] is None:
        raise StonerLoadError("Unable to find a delimiter in the line")
    return current[0]


@register_loader(
    patterns=[(".dat", 8), (".txt", 8), ("*", 8)],
    mime_types=[("application/tsv", 8), ("text/plain", 8), ("text/tab-separated-values", 8)],
    name="DataFile",
    what="Data",
)
def load_tdi_format(new_data, filename, *args, **kargs):
    """Actually load the data from disc assuming a .tdi file format.

    Args:
        filename (str):
            Path to filename to be loaded. If None or False, a dialog bax is raised to ask for the filename.

    Returns:
        DataFile:
            A copy of the newly loaded :py:class`DataFile` object.

    Exceptions:
        StonerLoadError:
            Raised if the first row does not start with 'TDI Format 1.5' or 'TDI Format=1.0'.

    Note:
        The *_load* methods shouldbe overridden in each child class to handle the process of loading data from
        disc. If they encounter unexpected data, then they should raise StonerLoadError to signal this, so that
        the loading class can try a different sub-class instead.
    """
    if filename is None or not filename:
        new_data.get_filename("r")
    else:
        new_data.filename = filename
    with FileManager(new_data.filename, "r", encoding="utf-8", errors="ignore") as datafile:
        line = datafile.readline()
        if line.startswith("TDI Format 1.5"):
            fmt = 1.5
        elif line.startswith("TDI Format=Text 1.0"):
            fmt = 1.0
        else:
            raise StonerLoadError("Not a TDI File")

        datafile.seek(0)
        reader = csv.reader(datafile, dialect=tab_delimited())
        cols = 0
        for ix, metadata in enumerate(reader):
            if ix == 0:
                row = metadata
                continue
            if len(metadata) < 1:
                continue
            if cols == 0:
                cols = len(metadata)
            if len(metadata) > 1:
                max_rows = ix + 1
            if "=" in metadata[0]:
                new_data.metadata.import_key(metadata[0])
        col_headers_tmp = [x.strip() for x in row[1:]]
        with warnings.catch_warnings():
            datafile.seek(0)
            warnings.filterwarnings("ignore", "Some errors were detected !")
            data = np.genfromtxt(
                datafile,
                skip_header=1,
                usemask=True,
                delimiter="\t",
                usecols=range(1, cols),
                invalid_raise=False,
                comments="\0",
                missing_values=[""],
                filling_values=[np.nan],
                max_rows=max_rows,
            )
    if data.ndim < 2:
        data = np.ma.atleast_2d(data)
    retain = np.all(np.isnan(data), axis=1)
    new_data.data = DataArray(data[~retain])
    new_data["TDI Format"] = fmt
    new_data["Stoner.class"] = "Data"
    if new_data.data.ndim == 2 and new_data.data.shape[1] > 0:
        new_data.column_headers = col_headers_tmp
    new_data.metadata = copy.deepcopy(new_data.metadata)  # This fixes some type issues TODO - work out why!
    return new_data


@register_saver(
    patterns=[(".dat", 8), (".txt", 8), ("*", 8)],
    name="DataFile",
    what="Data",
)
def save_tdi_format(save_data, filename):
    """Write out a DataFile to a tab delimited tdi text file."""

    header = ["TDI Format 1.5"]
    header.extend(save_data.column_headers[: save_data.data.shape[1]])
    header = "\t".join(header)
    mdkeys = sorted(save_data.metadata)
    if len(mdkeys) > len(save_data):
        mdremains = mdkeys[len(save_data) :]
        mdkeys = mdkeys[0 : len(save_data)]
    else:
        mdremains = []
    mdtext = np.array([save_data.metadata.export(k) for k in mdkeys])
    if len(mdtext) < len(save_data):
        mdtext = np.append(mdtext, np.zeros(len(save_data) - len(mdtext), dtype=str))
    data_out = np.column_stack([mdtext, save_data.data])
    fmt = ["%s"] * data_out.shape[1]
    with io.open(filename, "w", errors="replace", encoding="utf-8") as f:
        np.savetxt(f, data_out, fmt=fmt, header=header, delimiter="\t", comments="")
        for k in mdremains:
            f.write(save_data.metadata.export(k) + "\n")  # (str2bytes(save_data.metadata.export(k) + "\n"))

    save_data.filename = filename
    return save_data


@register_loader(
    patterns=[(".csv", 32), (".txt", 256)],
    mime_types=[("application/csv", 16), ("text/plain", 256), ("text/csv", 16)],
    name="CSVFile",
    what="Data",
)
def load_csvfile(new_data: "DataFile", filename: str, *args, **kargs) -> "DataFile":
    """Load generic deliminated files.

    Args:
        filename (string or bool): File to load. If None then the existing filename is used,
            if False, then a file dialog will be used.

    Keyword Arguments:
        header_line (int): The line in the file that contains the column headers.
            If None, then column headers are automatically generated.
        data_line (int): The line on which the data starts
        data_delim (string): The delimiter used for separating data values
        header_delim (strong): The delimiter used for separating header values

    Returns:
        A copy of the current object after loading the data.
    """
    _defaults = {"header_line": 0, "data_line": 1, "header_delim": ",", "data_delim": ","}

    header_line = kargs.pop("header_line", _defaults["header_line"])
    data_line = kargs.pop("data_line", _defaults["data_line"])
    data_delim = kargs.pop("data_delim", _defaults["data_delim"])
    header_delim = kargs.pop("header_delim", _defaults["header_delim"])

    new_data.filename = filename

    if data_delim is None or header_delim is None:  #
        with FileManager(new_data.filename, "r") as datafile:
            lines = datafile.readlines()
        if header_line is not None and header_delim is None:
            header_delim = _delim_detect(lines[header_line])
        if data_line is not None and data_delim is None:
            data_delim = _delim_detect(lines[data_line])

    with FileManager(new_data.filename, "r") as datafile:
        if header_line is not None:
            try:
                for ix, line in enumerate(datafile):
                    if ix == header_line:
                        break
                else:
                    raise StonerLoadError("Ran out of file before readching header")
                header = line.strip()
                column_headers = next(csv.reader(io.StringIO(header), delimiter=header_delim))
                data = np.genfromtxt(datafile, delimiter=data_delim, skip_header=data_line - header_line)
            except (TypeError, ValueError, csv.Error, StopIteration, UnicodeDecodeError) as err:
                raise StonerLoadError("Header and data on the same line") from err
        else:  # Generate
            try:
                data = np.genfromtxt(datafile, delimiter=data_delim, skip_header=data_line)
            except (TypeError, ValueError) as err:
                raise StonerLoadError("Failed to open file as CSV File") from err
            column_headers = ["Column" + str(x) for x in range(np.shape(data)[1])]

    new_data.data = data
    new_data.column_headers = column_headers
    new_data.metadata |= kargs
    return new_data


@register_saver(patterns=[(".csv", 32), (".txt", 256)], name="CSVFile", what="Data")
def save_csvfile(save_data: "DataFile", filename: Optional[str] = None, **kargs):
    """Override the save method to allow CSVFiles to be written out to disc (as a mininmalist output).

    Args:
        filename (string): Filename to save as (using the same rules as for the load routines)

    Keyword Arguments:
        deliminator (string): Record deliniminator (defaults to a comma)
        no_header (bool): Whether to skip the headers, defaults to False (include colu,n headers)

    Returns:
        A copy of itsave_data.
    """
    delimiter = kargs.pop("deliminator", ",")
    if filename is None:
        filename = save_data.filename
    no_header = kargs.get("no_header", False)
    if filename is None or (isinstance(filename, bool) and not filename):  # now go and ask for one
        filename = save_data.__file_dialog("w")
    with FileManager(filename, "w") as outfile:
        spamWriter = csv.writer(outfile, delimiter=delimiter, quotechar='"', quoting=csv.QUOTE_MINIMAL)
        i = 0
        if not no_header:
            spamWriter.writerow(save_data.column_headers)
        while i < save_data.data.shape[0]:
            spamWriter.writerow(save_data.data[i, :])
            i += 1
    save_data.filename = filename
    return save_data


@register_loader(
    patterns=[(".csv", 64), (".txt", 512)],
    mime_types=[("application/csv", 64), ("text/plain", 512), ("text/csv", 64)],
    name="JustNumbers",
    what="Data",
)
def load_justnumbers(new_data: "DataFile", filename: str, *args, **kargs) -> "DataFile":
    """Load generic deliminated files with no headers or metadata.

    Args:
        filename (str):
            File to load.

    Keyword Arguments:
        data_delim (string): The delimiter used for separating data values
        header_delim (strong): The delimiter used for separating header values

    Returns:
        A copy of the current object after loading the data.
    """
    _defaults = {"header_line": None, "data_line": 0, "data_delim": None}
    for karg, val in _defaults.items():
        kargs.setdefault(karg, val)
    return load_csvfile(new_data, filename, *args, **kargs)


@register_saver(patterns=[(".csv", 32), (".txt", 256)], name="JustNumbersFile", what="Data")
def save_justnumbers(save_data: "DataFile", filename: Optional[str] = None, **kargs):
    """Override the save method to allow JustNumbersFiles to be written out to disc (as a very mininmalist output).

    Args:
        filename (string): Filename to save as (using the same rules as for the load routines)

    Keyword Arguments:
        deliminator (string): Record deliniminator (defaults to a comma)

    Returns:
        A copy of itsave_data.
    """
    kargs["no_header"] = False
    return save_csvfile(save_data, filename, **kargs)


def _check_png_signature(filename):
    """Check that this is a PNG file and raie a StonerLoadError if not."""
    try:
        with FileManager(filename, "rb") as test:
            sig = test.read(8)
        sig = [x for x in sig]
        if sig != [137, 80, 78, 71, 13, 10, 26, 10]:
            raise StonerLoadError("Signature mismatrch")
    except (StonerLoadError, IOError) as err:
        from traceback import format_exc

        raise StonerLoadError(f"Not a PNG file!>\n{format_exc()}") from err
    return True


@register_loader(
    patterns=(".png", 16),
    mime_types=("image/png", 16),
    name="KermitPNGFile",
    what="Data",
)
def load_pngfile(new_data, filename=None, *args, **kargs):
    """PNG file loader routine.

    Args:
        filename (string or bool): File to load. If None then the existing filename is used,
            if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    new_data.filename = filename
    _check_png_signature(filename)
    try:
        with PIL.Image.open(new_data.filename, "r") as img:
            for k in img.info:
                new_data.metadata[k] = img.info[k]
            new_data.data = np.asarray(img)
    except IOError as err:
        raise StonerLoadError("Unable to read as a PNG file.") from err
    new_data["Stoner.class"] = "Data"  # Reset my load class
    new_data.metadata = copy.deepcopy(new_data.metadata)  # Fixes up some data types (see TDI loader)
    return new_data


@register_saver(patterns=(".png", 8), name="ImageFile", what="Image")
@register_saver(patterns=(".png", 16), name="KermitPngFile", what="Data")
def save_pngfile(save_data, filename=None, **kargs):
    """Override the save method to allow KermitPNGFiles to be written out to disc.

    Args:
        filename (string): Filename to save as (using the same rules as for the load routines)

    Keyword Arguments:
        deliminator (string): Record deliniminator (defaults to a comma)

    Returns:
        A copy of itsave_data.
    """
    if filename is None:
        filename = save_data.filename
    if filename is None or (isinstance(filename, bool) and not filename):  # now go and ask for one
        filename = save_data.__file_dialog("w")

    metadata = PIL.PngImagePlugin.PngInfo()
    for k in save_data.metadata:
        parts = save_data.metadata.export(k).split("=")
        key = parts[0]
        val = str2bytes("=".join(parts[1:]))
        metadata.add_text(key, val)
    img = PIL.Image.fromarray(save_data.data)
    img.save(filename, "png", pnginfo=metadata)
    save_data.filename = filename
    return save_data


try:  # Optional tdms support
    from nptdms import TdmsFile

    @register_loader(
        patterns=[(".tdms", 16), (".tdms_index", 16)],
        mime_types=("application/octet-stream", 16),
        name="TDMSFile",
        what="Data",
    )
    def load_tdms(new_data, filename=None, *args, **kargs):
        """TDMS file loader routine.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itnew_data after loading the data.
        """
        filename = Path(filename)
        if filename.suffix == ".tdms_index":
            filename = filename.parent / f"{filename.stem}.tdms"  # rewrite filename for not the index file!
        new_data.filename = filename
        # Open the file and read the main file header and unpack into a dict
        try:
            f = TdmsFile(new_data.filename)
            for grp in f.groups():
                if grp.path == "/":
                    pass  # skip the rooot group
                elif grp.path == "/'TDI Format 1.5'":
                    tmp = make_Data(grp.as_dataframe())
                    new_data.data = tmp.data
                    new_data.column_headers = tmp.column_headers
                    new_data.metadata.update(grp.properties)
                else:
                    tmp = make_Data(grp.as_dataframe())
                    new_data.data = tmp.data
                    new_data.column_headers = tmp.column_headers
        except (IOError, ValueError, TypeError, StonerLoadError) as err:
            from traceback import format_exc

            raise StonerLoadError(f"Not a TDMS File \n{format_exc()}") from err

        return new_data

except ImportError:
    pass

if Hyperspy_ok:

    def _unpack_hyperspy_meta(new_data, root, value):
        """Recursively unpack a nested dict of metadata and append keys to new_data.metadata."""
        if isinstance(value, Mapping):
            for item in value.keys():
                if root != "":
                    _unpack_hyperspy_meta(new_data, f"{root}.{item}", value[item])
                else:
                    _unpack_hyperspy_meta(new_data, f"{item}", value[item])
        else:
            new_data.metadata[root] = value

    def _unpack_hyperspy_axes(newdata, ax_manager):
        """Unpack the axes managber as metadata."""
        _axes_keys = ["name", "scale", "low_index", "low_value", "high_index", "high_value"]
        for ax in ax_manager.signal_axes:
            for k in _axes_keys:
                newdata.metadata[f"{ax.name}.{k}"] = getattr(ax, k)

    @register_loader(
        patterns=[(".emd", 32), (".dm4", 32)],
        mime_types=[("application/x-hdf", 64), ("application/x-hdf5", 64)],
        name="HyperSpyFile",
        what="Data",
    )
    def hypersput_load(new_data, filename=None, *args, **kargs):
        """Load HyperSpy file loader routine.

        Args:
            filename (string or bool): File to load. If None then the existing filename is used,
                if False, then a file dialog will be used.

        Returns:
            A copy of the itnew_data after loading the data.
        """
        if filename is None or not filename:
            new_data.get_filename("r")
        else:
            new_data.filename = filename
        # Open the file and read the main file header and unpack into a dict
        try:
            with catch_sysout():
                signal = hsload(new_data.filename)
            if hasattr(hs, "signals"):
                Signal2D = hs.signals.Signal2D
            else:
                Signal2D = hs.api.signals.Signal2D
            if not isinstance(signal, Signal2D):
                raise StonerLoadError("Not a 2D signal object - aborting!")
        except Exception as err:  # pylint: disable=W0703 Pretty generic error catcher
            raise StonerLoadError(f"Not readable by HyperSpy error was {err}") from err
        new_data.data = signal.data
        _unpack_hyperspy_meta(new_data, "", signal.metadata.as_dictionary())
        _unpack_hyperspy_axes(new_data, signal.axes_manager)
        del new_data["General.FileIO.0.timestamp"]
        return new_data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Feature Finding functions for analysis code."""

__all__ = ["FeatureOpsMixin"]

import numpy as np
from scipy.signal import find_peaks
from scipy.interpolate import interp1d

from Stoner.tools import isiterable, isTuple
from Stoner.core.exceptions import assertion
from .utils import threshold


class FeatureOpsMixin:
    """Mixin to provide additional functions to support finding features in a dataset."""

    def peaks(self, **kargs):
        """Locates peaks and/or troughs in a column of data by using SG-differentiation.

        Args:
            ycol (index):
                the column name or index of the data in which to search for peaks
            width (int or float):
                the expected minimum halalf-width of a peak in terms of the number of data points (int) or distance
                in x (float). This is used in the differnetiation code to find local maxima. Bigger equals less
                sensitive to experimental noise, smaller means better eable to see sharp peaks
            poly (int):
                the order of polynomial to use when differentiating the data to locate a peak. Must >=2, higher numbers
                will find sharper peaks more accurately but at the risk of finding more false positives.

        Keyword Arguments:
            significance (float):
                used to decide whether a local maxmima is a significant peak. Essentially just the curvature
                of the data. Bigger means less sensitive, smaller means more likely to detect noise. Default is the
                maximum curvature/(2*width)
            xcol (index or None):
                name or index of data column that p[provides the x-coordinate (default None)
            peaks (bool):
                select whether to measure peaks in data (default True)
            troughs (bool):
                select whether to measure troughs in data (default False)
            sort (bool):
                Sor the results by significance of peak
            modify (book):
                If true, then the returned object is a copy of self with only the peaks/troughs left in the data.
            full_data (bool):
                If True (default) then all columns of the data at which peaks in the *ycol* column are found.
                *modify* true implies *full_data* is also true. If *full_data* is False, then only the x-column
                values of the peaks are returned.

        Returns:
            (various):
                If *modify* is true, then returns a the AnalysisMixin with the data set to just the peaks/troughs.
                If *modify* is false (default), then the return value depends on *ycol* and *xcol*. If *ycol* is
                not None and *xcol* is None, then returns complete rows of data corresponding to the found
                peaks/troughs. If *xcol* is not None, or *ycol* is None and *xcol* is None, then returns a 1D array
                of the x positions of the peaks/troughs.

        See Also:
            User guide section :ref:`peak_finding`
        """
        width = kargs.pop("width", int(len(self) / 20))
        peaks = kargs.pop("peaks", True)
        troughs = kargs.pop("troughs", False)
        poly = kargs.pop("poly", 2)
        assertion(
            poly >= 2, "poly must be at least 2nd order in peaks for checking for significance of peak or through"
        )

        sort = kargs.pop("sort", False)
        modify = kargs.pop("modify", False)
        full_data = kargs.pop("full_data", True)
        _ = self._col_args(scalar=False, xcol=kargs.pop("xcol", None), ycol=kargs.pop("ycol", None))
        xcol, ycol = _.xcol, _.ycol
        if isiterable(ycol):
            ycol = ycol[0]
        if isinstance(width, float):  # Convert a floating point width unto an integer.
            xmin, xmax = self.span(xcol)
            width = int(len(self) * width / (xmax - xmin))
        width = max(width, poly + 1)
        setas = self.setas.clone  # pylint: disable=E0203
        self.setas = ""
        d1 = self.SG_Filter(ycol, xcol=xcol, points=width, poly=poly, order=1).ravel()
        d2 = self.SG_Filter(
            ycol, xcol=xcol, points=2 * width, poly=poly, order=2
        ).ravel()  # 2nd differential requires more smoothing

        # We're going to ignore the start and end of the arrays
        index_offset = int(width / 2)
        d1 = d1[index_offset:-index_offset]
        d2 = d2[index_offset:-index_offset]

        # Pad the ends of d2 with the mean value
        pad = np.mean(d2[index_offset:-index_offset])
        d2[:index_offset] = pad
        d2[-index_offset:] = pad

        # Set the significance from the 2nd ifferential if not already set
        significance = kargs.pop(
            "significance", np.max(np.abs(d2)) / (2 * width)
        )  # Base an apriori significance on max d2y/dx2 / 20
        if isinstance(significance, int):  # integer significance is inverse to floating
            significance = np.max(np.abs(d2)) / significance  # Base an apriori significance on max d2y/dx2 / 20

        d2_interp = interp1d(np.arange(len(d2)), d2, kind="cubic")
        # Ensure we have some X-data
        if xcol is None:
            xdata = np.arange(len(self))
        else:
            xdata = self.column(xcol)
        xdata = interp1d(np.arange(len(self)), xdata, kind="cubic")

        possible_peaks = np.array(threshold(0, d1, rising=troughs, falling=peaks))
        curvature = np.abs(d2_interp(possible_peaks))

        # Filter just the significant peaks
        possible_peaks = np.array([p for ix, p in enumerate(possible_peaks) if abs(curvature[ix]) > significance])
        # Sort in order of significance
        if sort:
            possible_peaks = np.take(possible_peaks, np.argsort(np.abs(d2_interp(possible_peaks))))

        xdat = xdata(possible_peaks + index_offset)

        if modify:
            self.data = self.interpolate(xdat, xcol=xcol, kind="cubic")
            ret = self
        elif full_data:
            ret = self.interpolate(xdat, kind="cubic", xcol=False)
        else:
            ret = xdat
        self.setas = setas
        # Return - but remembering to add back on the offset that we took off due to differentials not working at
        # start and end
        return ret

    def find_peaks(self, **kargs):
        """Interface to :py:func:`scipy.signal.find_peaks` for locating peaks in data.

        Args:
            ycol (index):
                the column name or index of the data in which to search for peaks

        Keyword Arguments:
            xcol (index):
                the column name or index of the x data that the peaks correspond to.
            height : number or ndarray or sequence, optional
                Required height of peaks. Either a number, ``None``, an array matching
                `ycol` or a 2-element sequence of the former. The first element is
                always interpreted as the  minimal and the second, if supplied, as the
                maximal required height.
            threshold : number or ndarray or sequence, optional
                Required threshold of peaks, the vertical distance to its neighbouring
                samples. Either a number, ``None``, an array matching `ycol` or a
                2-element sequence of the former. The first element is always
                interpreted as the  minimal and the second, if supplied, as the maximal
                required threshold.
            distance : number, optional
                Required minimal horizontal distance (>= 1) in samples between
                neighbouring peaks. Smaller peaks are removed first until the condition
                is fulfilled for all remaining peaks. If this is a *float* and *xcol* is set, then
                the units are in terms of the x-data, otherwise in rwo indices.
            prominence : number or ndarray or sequence, optional
                Required prominence of peaks. Either a number, ``None``, an array
                matching `ycol` or a 2-element sequence of the former. The first
                element is always interpreted as the  minimal and the second, if
                supplied, as the maximal required prominence.
            width : number or ndarray or sequence, optional
                Required width of peaks in samples. Either a number, ``None``, an array
                matching `ycol` or a 2-element sequence of the former. The first
                element is always interpreted as the  minimal and the second, if
                supplied, as the maximal required width. If this is a *float* and *xcol* is set, then
                the units are in terms of the x-data, otherwise in rwo indices.
            wlen : int, optional
                Used for calculation of the peaks prominences, thus it is only used if
                one of the arguments `prominence` or `width` is given. See argument
                `wlen` in `peak_prominences` for a full description of its effects.
            rel_height : float, optional
                Used for calculation of the peaks width, thus it is only used if `width`
                is given. See argument  `rel_height` in `peak_widths` for a full
                description of its effects.
            plateau_size : number or ndarray or sequence, optional
                Required size of the flat top of peaks in samples. Either a number,
                ``None``, an array matching `ycol` or a 2-element sequence of the former.
                The first element is always interpreted as the minimal and the second,
                if supplied as the maximal required plateau size. If this is a *float* and *xcol* is set, then
                the units are in terms of the x-data, otherwise in rwo indices.
            prefix (str):
                If et, then the metadata keys that return information about the peaks is returned with the given
                prefix. Default is None - no prefix.
            sort (bool):
                Sor the results by prominence of peak
            modify (book):
                If true, then the returned object is a copy of self with only the peaks left in the data.
            full_data (bool):
                If True (default) then all columns of the data at which peaks in the *ycol* column are found.
                *modify* true implies *full_data* is also true. If *full_data* is False, then only the x-column
                values of the peaks are returned.

        Returns:
            (various):
                If *modify* is true, then returns a the AnalysisMixin with the data set to just the peaks/troughs.
                If *modify* is false (default), then the return value depends on *ycol* and *xcol*. If *ycol* is
                not None and *xcol* is None, then returns complete rows of data corresponding to the found
                peaks/troughs. If *xcol* is not None, or *ycol* is None and *xcol* is None, then returns a 1D array
                of the x positions of the peaks/troughs.

        See Also:
            User guide section :ref:`peak_finding`
        """
        distance = kargs.pop("distance", None)
        width = kargs.pop("width", None)
        plateau_size = kargs.pop("plateau_size", None)
        sort = kargs.pop("sort", False)
        modify = kargs.pop("modify", False)
        bounds = kargs.pop("bounds", lambda x, y: True)
        prefix = kargs.pop("prefix", None)
        full_data = kargs.pop("full_data", True)
        _ = self._col_args(scalar=False, xcol=kargs.pop("xcol", None), ycol=kargs.pop("ycol", None))
        xcol, ycol = _.xcol, _.ycol
        if isiterable(ycol):
            ycol = ycol[0]

        if isinstance(width, float):  # Convert a floating point width unto an integer.
            xmin, xmax = self.span(xcol)
            width = int(len(self) * width / (xmax - xmin))
        elif isTuple(width, float, float):
            xmin, xmax = self.span(xcol)
            width = int(len(self) * width[0] / (xmax - xmin)), int(len(self) * width[1] / (xmax - xmin))
        if width is not None:
            kargs["width"] = width

        if isinstance(distance, float):  # Convert a floating point width unto an integer.
            xmin, xmax = self.span(xcol)
            distance = int(np.ceil(len(self) * distance / (xmax - xmin)))
        if distance is not None:
            kargs["distance"] = distance

        if isinstance(plateau_size, float):  # Convert a floating point plateau_size unto an integer.
            xmin, xmax = self.span(xcol)
            plateau_size = int(len(self) * plateau_size / (xmax - xmin))
        elif isTuple(plateau_size, float, float):
            xmin, xmax = self.span(xcol)
            plateau_size = (
                int(len(self) * plateau_size[0] / (xmax - xmin)),
                int(len(self) * plateau_size[1] / (xmax - xmin)),
            )
        if plateau_size is not None:
            kargs["plateau_size"] = plateau_size

        seek = self.search(xcol, bounds)
        peaks, data = find_peaks(seek[:, ycol], **kargs)
        peaks = self.data.i[seek.i[peaks]]  # de-reference frombounded data back to main dataset

        for sort_key in ["prominences", "peak_heights", "widths"]:
            if sort_key in data:
                break
        else:
            sort_key = None
        if sort and sort_key:
            idx = np.sort(np.array(list(zip(data[sort_key], np.arange(peaks.size)))), axis=0)[:, 1].astype(int)
            peaks = peaks[idx]
            for k in data:
                data[k] = data[k][idx]
        xmin, xmax = self.span(_.xcol)
        xconv = len(self) / (xmax - xmin)
        for k, v in data.items():
            if k.startswith("left") or k.startswith("right") or k == "widths":
                data[k] = v / xconv + (xmin if k != "widths" else 0)
        peak_data = self.data[peaks, :]
        for k, v in data.items():
            if prefix is None:
                self[k] = v
            else:
                self[f"{prefix}:{k}"] = v
        if modify:
            self.data = peak_data
            return self
        if full_data:
            return peak_data, data
        return peak_data[_.xcol], peak_data[_.yxol], data



# -*- coding: utf-8 -*-
"""Test for Stoner.analysis.filtering"""

import pytest
import numpy as np
from Stoner import Data
import warnings

warnings.filterwarnings("error")

testd = None
np.random.seed(12345)


def setup_function(function):
    global testd
    x = np.linspace(0, 2 * np.pi, 721)
    np.random.seed(87654321)
    y = np.sin(2 * x) + np.random.normal(scale=0.02, size=x.size)
    testd = Data(x, y, column_headers=["Time", "Ramp", "sine"])


def test_sgfilter():
    global testd
    testd.setas = ".y"
    testd.SG_Filter(result=True, header="Smooth")
    testd.setas = "xyy"
    res = testd // 1 - testd // 2
    assert res.mean() < 1e-4
    testd.del_column(2)
    testd.setas = ".y"
    testd.SG_Filter(result=True, header="Smooth", pad=0.0)
    testd.setas = "xyy"
    res = testd // 1 - testd // 2
    assert res.mean() < 1e-4
    testd.del_column(2)
    testd.SG_Filter(order=1, result=True, header="cos", pad=False)
    testd.setas = "xyz"
    testd.z /= 2
    testd.add_column(testd.z**2 + testd.y**2, header="sin*cos")
    testd.setas = "xyyy"
    testd.SG_Filter(col=3, order=0, points=51, replace=True, result=True, pad=1.0)
    testd.del_column(3)
    testd.setas = "xyyy"
    assert np.abs(testd // 3 - 1.0).max() < 0.075


def test_extrapolate():
    global testd
    testd.setas = "xy"
    testd.x -= 3
    with pytest.raises(TypeError):
        testd.extrapolate(kind=3.0)
    with pytest.raises(TypeError):
        testd.extrapolate(kind=lambda x, m, c: m * x + c)


def test_bins():
    global testd
    with pytest.raises(ValueError):
        testd.make_bins(0, 10, mode="bad")
    with pytest.raises(ValueError):
        testd.make_bins(0, 10.0, mode="log")
    with pytest.raises(ValueError):
        testd.make_bins(0, 0.5, mode="log")
    with pytest.raises(ValueError):
        testd.make_bins(0, 0.5, mode="bad")
    bins = np.linspace(testd.x.min(), testd.x.max(), 11)
    with pytest.raises(ValueError):
        testd.make_bins(0, bins=bins, mode="bad")

    b1 = testd.make_bins(0, bins=bins, mode="lin")
    for element in b1:
        assert len(element) == 10
    b2 = testd.make_bins(0, bins=bins, mode="log")
    for element in b2:
        assert len(element) == 10
    with pytest.raises(TypeError):
        testd.make_bins(0, "10", mode="bad")
    with pytest.raises(ValueError):
        testd.make_bins(0, np.linspace(0, 6, 1000), mode="lin")


def test_outlier_detect():
    global testd
    testd.add_column(np.zeros_like(testd.x), header="zeros")
    testd.data[[90, 270, 450, 630], 1] = 0
    d1 = testd.clone
    with pytest.raises(ValueError):
        d1.outlier_detection(action="bad")

    d1.outlier_detection(certainty=20)
    assert d1.count() in (717, 716)
    assert not np.any(d1.mask[:, 2])
    d1 = testd.clone
    d1.outlier_detection(certainty=20, action="mask row")
    assert d1.count() == 717
    assert np.all(d1.mask[90, :] == [True, True, True])
    d1 = testd.clone
    d1.outlier_detection(certainty=20, action="delete")
    assert len(d1) in (717, 716)
    with pytest.raises(SyntaxError):
        d1.outlier_detection(action_args=(True, False))

    def action(i, column, data):
        data[i, column] = (data[i - 1, column] + data[i + 1, column]) / 2

    d1 = testd.clone
    d1.outlier_detection(certainty=20, action=action)
    d1.outlier_detection(certainty=20, action="mask row")
    assert not np.any(d1.mask)


if __name__ == "__main__":
    pytest.main(["--pdb", __file__])



"""Create a 2D vector field plot."""

# pylint: disable=invalid-name
from os import path

from Stoner import Data, __home__

d = Data(path.join(__home__, "..", "sample-data", "OVF1.ovf"))
e = (
    d.select(Z__between=(10, 11))
    .select(X__between=(10, 18))
    .select(Y__between=(5, 13))
)
e.figure(figsize=(8, 4), no_axes=True)

# 2D vectors on a 2D Field
e.setas = "xy.uv."
e.subplot(121)
e.plot()
e.title = "3D Vector, 2D Field"

# 3D Vector on a 2D Field
e.subplot(122)
e.setas = "xy.uvw"
e.plot()
e.title = "3D Vector, 3D Field"



# -*- coding: utf-8 -*-
"""Test the ZIP module for zipfile compressed objects."""


import pytest
import os.path as path
import tempfile
import Stoner
import Stoner.Zip as SZ
import zipfile as zf

pth = path.dirname(__file__)

root = path.realpath(path.join(Stoner.__home__, ".."))
sample_data = path.realpath(path.join(root, "sample-data", "NLIV"))
tmpdir = tempfile.mkdtemp()


def test_zipFile(tmpdir):
    d = Stoner.Data(Stoner.__datapath__ / "TDI_Format_RT.txt")
    z = SZ.ZippedFile(d)
    z.save(path.join(tmpdir, "TDI_Format_RT.zip"))

    z2 = Stoner.Data(path.join(tmpdir, "TDI_Format_RT.zip"))
    z["Loaded as"] = z2["Loaded as"]
    z["Stoner.class"] = z2["Stoner.class"]
    assert z == z2
    with zf.ZipFile(path.join(tmpdir, "TDI_Format_RT.zip"), "r") as open_zipfile:
        z3 = Stoner.Data().load(open_zipfile, filetype=SZ.ZippedFile)
    z3["Stoner.class"] = z2["Stoner.class"]
    assert z2 == z3


def test_zipfolder():
    # Test constructor from DataFolder
    sf = Stoner.DataFolder(sample_data, pattern="*.txt")
    szf = SZ.ZipFolder(sf)
    assert sf.shape == szf.shape, "ZipFolder created from DataFolder didn't keep the same shape"
    assert sf[0] == szf[0], "First element of ZipFolder created from DataFolder changed!"
    zipname = path.join(tmpdir, "test-zipfolder.zip")
    szf.save(zipname)
    assert sf.shape == szf.shape, "ZipFolder Changed shape when saving!"
    szf_2 = SZ.ZipFolder(zipname).compress()
    assert szf_2.shape == szf.shape, "ZipFolder loaded from disc not same shape as ZipFolder in memory!"
    fname = path.basename(szf[0].filename)
    assert szf[fname] == szf_2[fname], "File from loaded ZipFolder not the same as in memory ZipFolder."


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



"""Example customising a plot using default style."""

# pylint: disable=invalid-name
import os.path as path
from cycler import cycler

from Stoner import Data, __home__
from Stoner.plot.formats import DefaultPlotStyle

filename = path.realpath(
    path.join(__home__, "..", "doc", "samples", "sample.txt")
)

d = Data(filename, setas="xyy", template=DefaultPlotStyle)
d.normalise(2, scale=(0.0, 10.0))  # Just to keep the plot sane!

# Change the color and line style cycles
d.template["axes.prop_cycle"] = cycler(color=["Blue", "Purple"]) + cycler(
    linestyle=["-", "--"]
)

# Can also access as an attribute
d.template.template_lines__linewidth = 2.0

# Set the default figure size
d.template["figure.figsize"] = (6, 8)
d.template["figure.autolayout"] = True
# Make figure (before using subplot method) and select first subplot
d.figure(no_axes=True)
d.subplot(212)
# Pkot with our customised defaults
d.plot()
d.grid(True, color="green", linestyle="-.")
d.title = "Customised Plot settings"
# Reset the template to defaults and switch to next subplot
d.template.clear()
d.subplot(211)
# Plot with defaults
d.plot()
d.title = "Style Default settings"
# Fixup layout
d.figwidth = 7  # Magic pass through attribute access



"""Classes and support functions for the :py:attr:`Stoner.DataFolder.each`.magic attribute."""

__all__ = ["Item"]
from collections.abc import MutableSequence
from functools import wraps, partial
from traceback import format_exc

import numpy as np

from ..tools import isiterable
from ..compat import string_types
from .utils import get_pool


def _worker(d, **kwargs):
    """Support function to run an arbitrary function over a :py:class:`Stoner.Data` object."""
    byname = kwargs.get("byname", False)
    func = kwargs.get("func", lambda x: x)
    if byname:
        func = getattr(d, func, lambda x: x)
    args = kwargs.get("args", tuple())
    kargs = kwargs.get("kargs", dict)
    if hasattr(d, "setas"):
        d["setas"] = list(d.setas)
    d["args"] = args
    d["kargs"] = kargs
    d["func"] = func.__name__
    try:
        if byname:  # Ut's an instance bound moethod
            ret = func(*args, **kargs)
        else:  # It's an arbitrary function
            ret = func(d, *args, **kargs)
    except Exception as e:  # pylint: disable=W0703 # Ok to be broad as user func could do anything
        ret = e, format_exc()
    return (d, ret)


class SetasWrapper(MutableSequence):
    """Manages wrapping each member of the folder's setas attribute."""

    def __init__(self, parent):
        """Note a reference to the parent item class instance and folder."""
        self._each = parent
        self._folder = parent._folder

    def __call__(self, *args, **kargs):
        """Pass through the calls the setas method of each item in our folder."""
        _ = self._folder._object_attrs.pop("setas", None)
        for obj in self._folder:
            obj.setas(*args, **kargs)
        self._folder._object_attrs["setas"] = self.collapse()

        return self._folder

    def __len__(self):
        """Return the shortest length of all the setas elements in the folder."""
        lengths = np.array([len(data.setas) for data in self._folder])
        return abs(lengths.min())

    def __getitem__(self, index):
        """Get the corresponding item from all the setas items in the folder."""
        return [data.setas[index] for data in self._folder]

    def __setitem__(self, index, value):
        """Set the value of the specified item on the setas elements in the folder.

        Args:
            index (int):
                The column index of the individual setas attributes to change
            value (iterable):
                The set of values to apply to the setas attributes.

        Notes:
            If value has a length less than the data folder, then the final value is repeated to encompass the
            remaining elements.
        """
        if len(value) < len(self._folder):
            value = value + value[-1] * (len(self._folder) - len(value))
        v = "."
        for v, data in zip(value, self._folder):
            data.setas[index] = v
        setas = self._folder._object_attrs.get("setas", self.collapse())
        setas[index] = v
        self._folder._object_attrs["setas"] = setas

    def __delitem__(self, index):
        """Cannot delete items from the proxy setas object - so simply clear it instead."""
        for data in self._folder:
            data.setas[index] = "."
        setas = self._folder._object_attrs.get("setas", self.collapse())
        setas[index] = "."
        self._folder._object_attrs["setas"] = setas

    def insert(self, index, value):
        """Cannot insert items into the proxy setas object."""
        raise IndexError("Cannot insert into the objectFolder's setas - insdert into the objectFolder instead!")

    def collapse(self):
        """Collapse the setas into a single list if possible."""
        setas = []
        for v in self:
            if len(v):
                if np.unique(v).size == 1:
                    setas.append(v[0])
                else:
                    setas.append("-")
        return setas


class Item:
    """Provides a proxy object for accessing methods on the inividual members of a Folder.

    Notes:
        The purpose of this class is to allow it to be explicit that we're calling methods
        on the members of the folder rather than a collective method. This allows us to work
        around nameclashes.
    """

    _folder = None

    def __init__(self, folder):
        """Create the each proxy object.

        Notes the partent folder that created us."""
        self._folder = folder

    @property
    def setas(self):
        """Return a proxy object for manipulating all the setas objects in a folder."""
        return SetasWrapper(self)

    @setas.setter
    def setas(self, value):
        """Manipualte the setas property of all the objects in the folder."""
        setas = self.setas
        setas(value)
        self._folder._object_attrs["setas"] = setas.collapse()
        return setas

    def __call__(self, func, *args, **kargs):
        """Iterate over the baseFolder, calling func on each item.

        Args:
            func (callable, str):
                Either a callable object, or the name of a callable object (either method or global) that must take
                a metadataObject type instance as it's first argument.

        Keyword Args:
            _return (None, bool or str): Controls how the return value from *func* is added to the DataFolder

        Returns:
            A list of the results of evaluating *func* for each item in the folder.

        Notes:
            If *_return* is None and the return type of *func* is the same type as the :py:class:`baseFolder` is
            storing, then the return value replaces the original :py:class:`Stoner.Core.metadataobject` in the
            :py:class:`baseFolder`. If *_result* is True the return value is added to the
            :py:class:`Stoner.Core.metadataObject`'s metadata under the name of the function. If *_result* is a
            string. then return result is stored in the corresponding name.
        """
        # Just call the iter generator but assemble into a list.
        if isinstance(func, string_types) and "_byname" not in kargs:
            if func in globals() and callable(globals()[func]):
                func = globals()[func]
            else:
                func = getattr(self, func)
        return list(self.iter(func, *args, **kargs))

    def __dir__(self):
        """Return a list of the common set of attributes of the instances in the folder."""
        if self._folder and len(self._folder) != 0:
            res = set(dir(self._folder[0]))
        else:
            res = set()
        if self._folder and len(self._folder) > 0:
            for d in self._folder[1:]:
                res &= set(dir(d))
        return list(res)

    def __delattr__(self, name):
        """Handle removing an attribute from the folder, including proxied attributes."""
        if name in dir(self._folder.instance) or (
            len(self._folder) and hasattr(self._folder[0], name)
        ):  # This is an instance attribute
            # If we're tracking the object attributes and have a type set, then we can store this for adding to all
            # loaded objects on read.
            del self._folder._object_attrs[name]
            for _, this in self._folder.loaded:
                try:
                    delattr(this, name)
                except AttributeError:
                    pass
        elif name in self._folder._instance_attrs:
            del self._folder._instance_attrs[name]
        else:
            raise AttributeError(f"Unrecognised attribute {name}")

    def __getattr__(self, name):
        """Handle some special case attributes that provide alternative views of the objectFolder.

        Args:
            item (string): The attribute name being requested

        Returns:
            Depends on the attribute

        Notes:
            If *name* is not present on the empty member instance, then the first member of the folder is checked as
            well. This allows the attributes of a :py:class:`Stoner.Data` object that derive from the
            Lpy:attr:`Stoner.Data.setas` attribute (such as *.x*, *.y* or *.e* etc) can be accessed.
        """
        try:
            return super().__getattr__(name)
        except AttributeError:
            pass
        try:
            instance = self._folder.instance
            if callable(getattr(instance, name, None)):  # It's a method
                ret = self.__getattr_proxy(name)
            else:  # It's a static attribute
                if name in self._folder._object_attrs and len(self._folder) == 0:
                    ret = self._folder._object_attrs[name]
                elif len(self._folder):
                    ret = [(not hasattr(x, name), getattr(x, name, None)) for x in self._folder]
                    mask, values = zip(*ret)
                    lens = np.unique([len(v) if isiterable(v) else 0 for v in values])
                    if lens.size == 1:
                        ret = np.ma.MaskedArray(values)
                    else:
                        ret = np.ma.MaskedArray(values, dtype=object)
                    ret.mask = mask
                else:
                    ret = getattr(instance, name, None)
                if ret is None or (hasattr(ret, "mask") and np.all(ret.mask)):
                    raise AttributeError
        except AttributeError as err:  # Ok, pass back
            raise AttributeError(f"{name} is not an Attribute of {type(self)} or {type(instance)}") from err
        # except TypeError as err:  # Can be triggered if self.instance lacks the attribute
        #     if len(self._folder) and hasattr(self._folder[0], name):
        #         ret = [(not hasattr(x, name), getattr(x, name, None)) for x in self._folder]
        #         mask, values = zip(*ret)
        #         ret = np.ma.MaskedArray(values)
        #         ret.mask = mask
        #     else:
        #         raise err

        return ret

    def __setattr__(self, name, value):
        """Proxy call to set an attribute.

        Setting the attribute on .each sets it on all instantiated objects and in _object_attrs.

        Args:
            name(str): Attribute to set
            value (any): Value to set

        Notes:
            If *name* is not present on the empty member instance, then the first member of the folder is checked as
            well. This allows the attributes of a :py:class:`Stoner.Data` object that derive from the
            Lpy:attr:`Stoner.Data.setas` attribute (such as *.x*, *.y* or *.e* etc) can be accessed.

            If *value* is iterable and the same length as the folder, then each element in the folder is loaded and
            the corresponding element of *value* is assigned to the attribute of the member.
        """
        if hasattr(type(self), name) or name.startswith("_"):  # Handle setting our own attributes
            super().__setattr__(name, value)
        elif name in dir(self._folder.instance) or (
            len(self._folder) and hasattr(self._folder[0], name)
        ):  # This is an instance attribute
            if isiterable(value) and len(value) == len(self._folder):
                force_load = True
            else:
                force_load = False
                self._folder._object_attrs[name] = value  # Add to attributes to be set on load
                value = [value] * len(self._folder)
            for d, v in zip(self._folder.__names__(), value):  # And set on all instantiated objects
                if force_load or isinstance(self._folder.__getter__(d, instantiate=False), self._folder.type):
                    d = self._folder.__getter__(d)
                    setattr(d, name, v)
        else:
            raise AttributeError(f"Unknown attribute {name}")

    def __getattr_proxy(self, item):
        """Make a prpoxy call to access a method of the metadataObject like types.

        Args:
            item (string): Name of method of metadataObject class to be called

        Returns:
            Either a modified copy of this objectFolder or a list of return values
            from evaluating the method for each file in the Folder.
        """
        meth = getattr(self._folder.instance, item, None)

        @wraps(meth)
        def _wrapper_(*args, **kargs):
            """Wrap a call to the metadataObject type for magic method calling.

            Keyword Arguments:
                _return (index types or None): specify to store the return value in the individual object's metadata

            Note:
                This relies on being defined inside the enclosure of the objectFolder method
                so we have access to self and item
            """
            kargs["_byname"] = True
            return self(item, *args, **kargs)  # Develove to self.__call__ where we have multiprocess magic

        # Ok that's the wrapper function, now return  it for the user to mess around with.
        return _wrapper_

    def __rmatmul__(self, other):
        """Implement callable@DataFolder as a generic iterate a function over DataFolder members.

        Returns:
            An object that supports __call__ and knows about this DataFolder.
        """
        if not callable(other):
            return NotImplemented

        @wraps(other)
        def _wrapper_(*args, **kargs):
            """Wrap a call to the metadataObject type for magic method calling.

            Keyword Arguments:
                _return (index types or None): specify to store the return value in the individual object's metadata

            Note:
                This relies on being defined inside the enclosure of the objectFolder method
                so we have access to self and item
            """
            kargs["_byname"] = False  # Force the __call__ to use the callable function
            return self(other, *args, **kargs)  # Delegate to self.__call__ which has multiprocess magic.

        # Ok that's the wrapper function, now return  it for the user to mess around with.
        return _wrapper_

    def iter(self, func, *args, **kargs):
        """Iterate over the baseFolder, calling func on each item.

        Args:
            func (callable): A Callable object that must take a metadataObject type instance as it's first argument.

        Keyword Args:
            _return (None, bool or str): Controls how the return value from *func* is added to the DataFolder

        Returns:
            A list of the results of evaluating *func* for each item in the folder.

        Notes:
            If *_return* is None and the return type of *func* is the same type as the :py:class:`baseFolder` is
            storing, then the return value replaces the original :py:class:`Stoner.Core.metadataobject` in the
            :py:class:`baseFolder`. If *_result* is True the return value is added to the
            :py:class:`Stoner.Core.metadataObject`'s metadata under the name of the function. If *_result* is a
            string. then return result is stored in the corresponding name.
        """
        _return = kargs.pop("_return", None)
        _byname = kargs.pop("_byname", False)
        _serial = kargs.pop("_serial", False)
        self._folder.fetch()  # Prefetch thefolder in case we can do it in parallel
        p, imap = get_pool(_serial)
        for ix, (new_d, ret) in enumerate(
            imap(partial(_worker, func=func, args=args, kargs=kargs, byname=_byname), self._folder)
        ):
            if self._folder.debug:
                print(ix, type(ret))
            if isinstance(ret, self._folder._type) and _return is None:
                try:  # Check if ret has same data type, otherwise will not overwrite well
                    if ret.data.dtype != new_d.data.dtype:
                        continue
                    new_d = ret
                except AttributeError:
                    pass
            elif _return is not None:
                if isinstance(_return, bool) and _return:
                    _return = func.__name__
                new_d[_return] = ret
            name = self._folder.__names__()[ix]
            self._folder.__setter__(name, new_d)
            yield ret
        if p is not None:
            p.close()
            p.join()



"""Demo of Fitting a directory of files."""

# pylint: disable=invalid-name
from os.path import join
from matplotlib.pyplot import figure

from Stoner import __home__, DataFolder, Data
from Stoner.plot.formats import TexEngFormatter
from Stoner.analysis.fitting.models.generic import Quadratic

# Set up the directory with our data
datafiles = join(__home__, "..", "sample-data", "NLIV")

# DataFolder of our data files
fldr = DataFolder(datafiles, pattern="*.txt").sort("y")

fldr.setas = "yx"

# Another Data object to keep the results in
result = Data()

# Loop over the files in the DataFolder
for d in fldr:
    # Fit, returning results as a dictionary
    row = d.lmfit(Quadratic, output="dict")
    # Add Extra information from metadata
    row["Magnetic Field $\\mu_0 H (T)$"] = d["y"]
    # Add the new row of data to the result.
    result += row

# Set columns and set the plot axis formatting to use SI precfixes
result.setas(x="Field", y="b", e="d_b")
result.template.xformatter = TexEngFormatter
result.template.yformatter = TexEngFormatter

# Plot
result.plot(fmt="k.", capsize=2)

# An alternative way to run the Analysis - this time with
# an orthogonal didstance regression algorithm

# Run the fitt for each file in the fldr. Set the output to "data" to
# Have the amended results replace the existing data files
fldr.each.odr(Quadratic, output="data", result=True, header="fit", _serial=True)
fig = figure()
fldr.each.setas(Current="x", Voltage="y")
fldr.each.setas[3] = "y"


fldr.each.plot(fmt=["+", "-"], label="Field = {y}mT", figure=fig, _serial=True)
fldr[0].legend(ncol=2, fontsize="xx-small")
fldr[0].title = "Non-local IV Curves"

# Now take a slice through the metadata to get the files we want.
result_2 = fldr.metadata.slice(["y", Quadratic], output="Data")

# Set the columns assignments and plot
result_2.setas = "x..ye"
result_2.plot(fmt="r.", figure=result.fig, capsize=2)



"""Core support for working with collections of files in the :py:class:`Stoner.DataFolder`."""

__all__ = ["core", "each", "groups", "metadata", "mixins", "utils", "DataFolder", "PlotFolder"]

from . import mixins
from .mixins import DataFolder, PlotFolder
